#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -------------------------------------------------------------------
#   Filename:  event_handler.py
#   Purpose:   handling events in obspyDMT
#   Author:    Kasra Hosseini
#   Email:     kasra.hosseinizad@earth.ox.ac.uk
#   License:   GNU Lesser General Public License, Version 3
# -------------------------------------------------------------------

# -----------------------------------------------------------------------
# ----------------Import required Modules (Python and Obspy)-------------
# -----------------------------------------------------------------------
from __future__ import print_function
from collections import OrderedDict
import copy
from datetime import datetime, timedelta
import glob
import numpy as np
from obspy.core.event import Catalog
try:
    from obspy.core.event import readEvents
except Exception as e:
    from obspy import read_events as readEvents
from obspy.core import UTCDateTime
try:
    from obspy.geodetics import locations2degrees
except Exception as e:
    from obspy.core.util import locations2degrees
try:
    from obspy.clients.fdsn import Client as Client_fdsn
except Exception as e:
    from obspy.fdsn import Client as Client_fdsn
import os
import pickle
import sys
import time
try:
    from urllib2 import urlopen 
except ImportError:
    from urllib.request import urlopen
try:
    from urllib import urlencode as urlencodeparse
except ImportError:
    from urllib.parse import urlencode as urlencodeparse
from .input_handler import input_logger
from .utility_codes import locate

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

# ##################### get_time_window ######################################


def get_time_window(input_dics, request):
    """
    generating a list of requests for both event_based and continuous requests.
    :param input_dics:
    :param request:
    :return:
    """
    t_event_1 = time.time()

    # request can be 'event_based' or 'continuous'
    # events list contains all the information for requested time-window
    # Although we do not have any events in continuous requests,
    # it is still called as events.
    events = []
    events_qml = Catalog(events=[])
    try:
        if not input_dics['read_catalog']:
            events_local = read_info(input_dics)
            if input_dics['event_catalog'].lower() == 'local':
                if events_local == 'no_local':
                    print("[WARNING] no local event was found!")
                    if request == 'event_based':
                        print("[WARNING] use IRIS catalog instead!")
                        input_dics['event_catalog'] = 'IRIS'
                else:
                    events = copy.deepcopy(events_local)
        else:
            events_QML, ev_already_list = \
                read_local_events(input_dics['read_catalog'])
            if not ev_already_list:
                # no matter if list was passed or requested, sort catalogue,
                # plot events and proceed
                events_QML = sort_catalogue(events_QML)
                events = qml_to_event_list(events_QML)
            else:
                events = copy.deepcopy(events_QML)
                events_QML = False
            for i in range(len(events)):
                events[i]['t1'] = events[i]['datetime'] - input_dics['preset']
                events[i]['t2'] = events[i]['datetime'] + input_dics['offset']

        if events == 'no_local' or events == []:
            if request.lower() == 'event_based':
                events, events_qml = event_info(input_dics)
            elif request.lower() == 'continuous':
                events, events_qml = continuous_info(input_dics)

        if not input_dics['continuous']:
            remove_indx = []
            if input_dics['event_catalog'].lower() != 'local':
                if events_local != 'no_local':
                    for nei in range(len(events)):
                        for oei in range(len(events_local)):
                            if (events[nei]['event_id'] ==
                                    events_local[oei]['event_id']):
                                remove_indx.append(nei)
                                break
                    if len(remove_indx) > 0:
                        remove_indx.sort(reverse=True)
                        for ri in remove_indx:
                            del events[ri]

                    for oei in range(len(events_local)):
                        events.append(events_local[oei])

    except Exception as error:
        print('WARNING: %s' % error)
        return []

    if len(events) == 0:
        return []

    events2, row_format, header = output_shell_event(events, request)

    print('Number of events/intervals: %s' % len(events))
    print('Time for retrieving and saving the event info: %s' \
          % str(timedelta(seconds=round(float(time.time() - t_event_1)))))

    if input_dics['primary_mode'] in ['event_based', 'continuous',
                                      'meta_data']:
        # formatting output / check if directory exists
        eventpath = os.path.join(input_dics['datapath'])
        write_cat_logger(input_dics, eventpath, events, events_qml,
                         events2, row_format, header)
    return events

# ##################### read_info #####################################


def read_info(input_dics):
    """
    read already created events (for both event_based and continuous requests)
    :param input_dics:
    :return:
    """
    evs_info = locate(input_dics['datapath'], 'EVENTS-INFO', num_matches=1)
    if len(evs_info) == 0:
        return "no_local"
    if len(evs_info) > 1:
        print("[WARNING] Found two directories that have EVENTS-INFO. Continue with:")
        print(evs_info[0])
        print('\n')

    ev_info = evs_info[0]

    if not os.path.isfile(os.path.join(ev_info, 'event_list_pickle')):
        print("[WARNING] no 'event_list_pickle' was found at %s, "
              "try to re-generate it..." % ev_info)
        info_dirs = locate(input_dics['datapath'], 'info')
        event_list_regen = []
        for info_d in info_dirs:
            if os.path.isfile(os.path.join(info_d, 'event.pkl')):
                try:
                    ev_pkl_tmp = pickle.load(open(os.path.join(info_d, 'event.pkl'), 'r'))
                    # to check whether t1 and t2 variables are there:
                    test_t1 = ev_pkl_tmp['t1']
                    test_t2 = ev_pkl_tmp['t2']
                    event_list_regen.append(ev_pkl_tmp)
                except Exception as e:
                    print("ERROR: %s" % e)
        if len(event_list_regen) > 0:
            print("[WARNING] a new 'event_list_pickle' was generated with %s events." % len(event_list_regen))
            fio = open(os.path.join(ev_info, 'event_list_pickle'), 'w')
            pickle.dump(event_list_regen, fio, protocol=2)
        else:
            print("[WARNING] a new 'event_list_pickle' could NOT be generated.")
            return "no_local"

    fio = open(os.path.join(ev_info, 'event_list_pickle'), 'rb')
    events = pickle.load(fio)
    if input_dics['event_catalog'].lower() == 'local':
        print("\n=========================================================")
        print("use the local files:")
        print(os.path.join(ev_info, 'event_list_pickle'))
        print("=========================================================\n")
    return events

# ##################### event_info #####################################


def event_info(input_dics):
    """
    get event(s) info for event_based request
    :param input_dics:
    :return:
    """
    try:
        evlatmin = input_dics['evlatmin']
        evlatmax = input_dics['evlatmax']
        evlonmin = input_dics['evlonmin']
        evlonmax = input_dics['evlonmax']

        evlat = input_dics['evlat']
        evlon = input_dics['evlon']
        evradmin = input_dics['evradmin']
        evradmax = input_dics['evradmax']

        event_switch = 'fdsn'
        event_url = input_dics['event_catalog']
        if input_dics['read_catalog']:
            event_switch = 'local'
        event_fdsn_cat = None

        if event_url.lower() == 'gcmt_combo':
            event_switch = 'gcmt_combo'
        if event_url.lower() == 'neic_usgs':
            event_switch = 'neic_usgs'
        if event_url.lower() == 'isc':
            event_switch = 'isc_cat'

        print('\nEvent(s) are based on:\t%s' % input_dics['event_catalog'])

        if event_switch == 'fdsn':
            client_fdsn = Client_fdsn(base_url=event_url)
            events_QML = client_fdsn.get_events(
                minlatitude=evlatmin,
                maxlatitude=evlatmax,
                minlongitude=evlonmin,
                maxlongitude=evlonmax,
                latitude=evlat,
                longitude=evlon,
                minradius=evradmin,
                maxradius=evradmax,
                mindepth=input_dics['min_depth'],
                maxdepth=input_dics['max_depth'],
                starttime=input_dics['min_date'],
                endtime=input_dics['max_date'],
                minmagnitude=input_dics['min_mag'],
                maxmagnitude=input_dics['max_mag'],
                orderby='time-asc',
                catalog=event_fdsn_cat,
                magnitudetype=input_dics['mag_type'],
                includeallorigins=None,
                includeallmagnitudes=None,
                includearrivals=None,
                eventid=None,
                limit=None,
                offset=None,
                contributor=None,
                updatedafter=None)

        elif event_switch == 'gcmt_combo':
            events_QML = \
                gcmt_catalog(input_dics['min_date'],
                             input_dics['max_date'],
                             evlatmin, evlatmax, evlonmin, evlonmax,
                             evlat, evlon, evradmin, evradmax,
                             input_dics['min_depth'],
                             input_dics['max_depth'],
                             input_dics['min_mag'],
                             input_dics['max_mag'])

        elif event_switch == 'neic_usgs':
            events_QML = \
                neic_catalog_urllib(input_dics['min_date'],
                                    input_dics['max_date'],
                                    evlatmin, evlatmax, evlonmin, evlonmax,
                                    evlat, evlon, evradmin, evradmax,
                                    input_dics['min_depth'],
                                    input_dics['max_depth'],
                                    input_dics['min_mag'],
                                    input_dics['max_mag'])

        elif event_switch == 'isc_cat':
            events_QML = \
                isc_catalog(bot_lat=evlatmin, top_lat=evlatmax,
                            left_lon=evlonmin, right_lon=evlonmax,
                            ctr_lat=evlat, ctr_lon=evlon,
                            radius=evradmax,
                            start_time=input_dics['min_date'],
                            end_time=input_dics['max_date'],
                            min_dep=input_dics['min_depth'],
                            max_dep=input_dics['max_depth'],
                            min_mag=input_dics['min_mag'],
                            max_mag=input_dics['max_mag'],
                            mag_type=input_dics['mag_type'],
                            req_mag_agcy='Any',
                            rev_comp=input_dics['isc_catalog'])

        elif event_switch == 'local':
            events_QML = readEvents(input_dics['read_catalog'])

        else:
            sys.exit('[ERROR] %s is not supported'
                     % input_dics['event_catalog'])

        events_delete = []
        for i in range(len(events_QML)):
            try:
                if not hasattr(events_QML.events[i], 'preferred_mag'):
                    if not hasattr(events_QML.events[i], 'preferred_magnitude'):
                        events_QML.events[i].preferred_mag = events_QML.events[i].magnitudes[0].mag
                        events_QML.events[i].preferred_mag_type = events_QML.events[i].magnitudes[0].magnitude_type
                        events_QML.events[i].preferred_author = 'None'
                    else:
                        events_QML.events[i].preferred_mag = events_QML.events[i].preferred_magnitude().mag
                        events_QML.events[i].preferred_mag_type = \
                            events_QML.events[i].preferred_magnitude().magnitude_type
                        events_QML.events[i].preferred_author = 'None'
                if not hasattr(events_QML.events[i], 'preferred_author'):
                    events_QML.events[i].preferred_author = 'None'
            except Exception as err:
                #print("[WARNING] event: %s, error: %s" % (i, err))
                events_delete.append(i)

        events_delete.sort(reverse=True)
        for i in events_delete:
            del events_QML[i]

        # no matter if list was passed or requested, sort catalogue,
        # plot events and proceed
        events_QML = sort_catalogue(events_QML)
        events = qml_to_event_list(events_QML)

    except Exception as error:
        print(60*'-')
        print('[WARNING] %s' % error)
        print(60*'-')
        events = []
        events_QML = []

    for i in range(len(events)):
        events[i]['t1'] = events[i]['datetime'] - input_dics['preset']
        events[i]['t2'] = events[i]['datetime'] + input_dics['offset']

    return events, events_QML

# ##################### read_local_events #####################################


def read_local_events(catalog_add):
    """
    read local events
    :param catalog_add:
    :return:
    """
    try:
        events_QML = readEvents(catalog_add)
        return events_QML, False
    except Exception as error:
        try:
            ev_csv = np.loadtxt(catalog_add, delimiter=',',
                                comments='#', ndmin=2, dtype='object')
            events = []
            for i in range(np.shape(ev_csv)[0]):
                events.append(OrderedDict(
                    [('number', i+1),
                     ('latitude', eval(ev_csv[i][3])),
                     ('longitude', eval(ev_csv[i][4])),
                     ('depth', eval(ev_csv[i][5])),
                     ('datetime', UTCDateTime(ev_csv[i][2])),
                     ('magnitude', eval(ev_csv[i][6])),
                     ('magnitude_type', ev_csv[i][7]),
                     ('author', ev_csv[i][8]),
                     ('event_id', ev_csv[i][1]),
                     ('origin_id', -12345),
                     ('focal_mechanism', [eval(ev_csv[i][10]),
                                          eval(ev_csv[i][11]),
                                          eval(ev_csv[i][12]),
                                          eval(ev_csv[i][13]),
                                          eval(ev_csv[i][14]),
                                          eval(ev_csv[i][15])
                                          ]),
                     ('source_duration', [ev_csv[i][16], eval(ev_csv[i][17])]),
                     ('flynn_region', ev_csv[i][9])]))
            return events, True
        except Exception as error:
            print(60*'-')
            print('[WARNING] %s' % error)
            print(60*'-')
            return Catalog(events=[]), False

# ##################### qml_to_event_list #####################################


def qml_to_event_list(events_QML):
    """
    convert QML to event list
    :param events_QML:
    :return:
    """
    events = []
    for i in range(len(events_QML)):
        try:
            event_time = events_QML.events[i].preferred_origin().time or \
                         events_QML.events[i].origins[0].time
            event_time_month = '%02i' % int(event_time.month)
            event_time_day = '%02i' % int(event_time.day)
            event_time_hour = '%02i' % int(event_time.hour)
            event_time_minute = '%02i' % int(event_time.minute)
            event_time_second = '%02i' % int(event_time.second)

            if not hasattr(events_QML.events[i], 'preferred_mag'):
                events_QML.events[i].preferred_mag = \
                    events_QML.events[i].magnitudes[0].mag
                events_QML.events[i].preferred_mag_type = \
                    events_QML.events[i].magnitudes[0].magnitude_type
                events_QML.events[i].preferred_author = 'None'
            else:
                if not hasattr(events_QML.events[i], 'preferred_author'):
                    if events_QML.events[i].preferred_magnitude().creation_info:
                        events_QML.events[i].preferred_author = \
                            events_QML.events[i].preferred_magnitude().creation_info.author
                    elif events_QML.events[i].magnitudes[0].creation_info:
                        events_QML.events[i].preferred_author = \
                            events_QML.events[i].magnitudes[0].creation_info.author
        except Exception as error:
            print(error)
            continue
        try:
            if not events_QML.events[i].focal_mechanisms == []:
                if events_QML.events[i].preferred_focal_mechanism()['moment_tensor']['tensor']:
                    focal_mechanism = [
                        events_QML.events[i].preferred_focal_mechanism()
                        ['moment_tensor']['tensor']['m_rr'],
                        events_QML.events[i].preferred_focal_mechanism()
                        ['moment_tensor']['tensor']['m_tt'],
                        events_QML.events[i].preferred_focal_mechanism()
                        ['moment_tensor']['tensor']['m_pp'],
                        events_QML.events[i].preferred_focal_mechanism()
                        ['moment_tensor']['tensor']['m_rt'],
                        events_QML.events[i].preferred_focal_mechanism()
                        ['moment_tensor']['tensor']['m_rp'],
                        events_QML.events[i].preferred_focal_mechanism()
                        ['moment_tensor']['tensor']['m_tp']]
                else:
                    found_foc_mech = False
                    for foc_mech_qml in events_QML.events[i].focal_mechanisms:
                        if foc_mech_qml['moment_tensor']['tensor']:
                            focal_mechanism = [
                                foc_mech_qml['moment_tensor']['tensor']['m_rr'],
                                foc_mech_qml['moment_tensor']['tensor']['m_tt'],
                                foc_mech_qml['moment_tensor']['tensor']['m_pp'],
                                foc_mech_qml['moment_tensor']['tensor']['m_rt'],
                                foc_mech_qml['moment_tensor']['tensor']['m_rp'],
                                foc_mech_qml['moment_tensor']['tensor']['m_tp']
                            ]
                            found_foc_mech = True
                            break
                    if not found_foc_mech:
                        focal_mechanism = False
            else:
                focal_mechanism = False
        except AttributeError:
            print("[WARNING] focal_mechanism does not exist for " \
                  "event: %s -- set to False" % (i+1))
            focal_mechanism = False
        except TypeError:
            focal_mechanism = False
        except Exception as error:
            print(error)
            focal_mechanism = False

        try:
            if not events_QML.events[i].focal_mechanisms == []:
                source_duration = [
                    events_QML.events[i].preferred_focal_mechanism()
                    ['moment_tensor']['source_time_function']['type'],
                    events_QML.events[i].preferred_focal_mechanism()
                    ['moment_tensor']['source_time_function']
                    ['duration']]
                if not source_duration[1]:
                    source_duration = mag_duration(
                        mag=events_QML.events[i].preferred_mag)
            else:
                source_duration = mag_duration(
                    mag=events_QML.events[i].preferred_mag)
        except AttributeError:
            print("[WARNING] source duration does not exist for " \
                  "event: %s -- set to False" % (i+1))
            source_duration = False
        except TypeError:
            source_duration = False
        except Exception as error:
            print(error)
            source_duration = False

        try:
            events.append(OrderedDict(
                [('number', i+1),
                 ('latitude',
                  events_QML.events[i].preferred_origin().latitude or
                  events_QML.events[i].origins[0].latitude),
                 ('longitude',
                  events_QML.events[i].preferred_origin().longitude or
                  events_QML.events[i].origins[0].longitude),
                 ('depth',
                  events_QML.events[i].preferred_origin().depth/1000. or
                  events_QML.events[i].origins[0].depth/1000.),
                 ('datetime', event_time),
                 ('magnitude',
                  events_QML.events[i].preferred_mag),
                 ('magnitude_type',
                  events_QML.events[i].preferred_mag_type),
                 ('author',
                  events_QML.events[i].preferred_author),
                 ('event_id', str(event_time.year) +
                  event_time_month + event_time_day + '_' +
                  event_time_hour + event_time_minute +
                  event_time_second + '.a'),
                 ('origin_id', events_QML.events[i].preferred_origin_id or
                  events_QML.events[i].origins[0].resource_id.resource_id),
                 ('focal_mechanism', focal_mechanism),
                 ('source_duration', source_duration),
                 ('flynn_region', 'NAN'),
                 ]))
        except Exception as error:
            print(error)
            continue
    return events

# ##################### continuous_info #####################################


def continuous_info(input_dics):
    """
    get 'event(s)' info for continuous request
    :param input_dics:
    :return:
    """
    print('start identifying the intervals...', end='')
    m_date = UTCDateTime(input_dics['min_date'])
    M_date = UTCDateTime(input_dics['max_date'])
    t_cont = M_date - m_date

    if t_cont < 0:
        sys.exit("\n\n[ERROR] max_date is lower than min_date!")

    events = []
    if t_cont > input_dics['interval']:
        num_div = int(t_cont/input_dics['interval'])
        # residual time is: (has not been used here)
        # t_res = t_cont - num_div*input_dics['interval']
        for i in range(1, num_div+1):
            cont_dir_name = (len(str(num_div)) - len(str(i)))*'0' + str(i)
            events.append(OrderedDict(
                [('number', i),
                 ('latitude', -12345),
                 ('longitude', -12345),
                 ('depth', -12345),
                 ('datetime', m_date + (i-1)*input_dics['interval']),
                 ('magnitude', -12345),
                 ('magnitude_type', 'NAN'),
                 ('author', 'NAN'),
                 ('event_id', 'continuous' + cont_dir_name),
                 ('origin_id', -12345),
                 ('focal_mechanism', False),
                 ('source_duration', False),
                 ('flynn_region', 'NAN'),
                 ('t1', m_date + (i-1)*input_dics['interval'] -
                  input_dics['preset']),
                 ('t2', m_date + i*input_dics['interval'] +
                  input_dics['offset']),
                 ]))

        final_time = m_date + num_div*input_dics['interval'] + \
            input_dics['offset']

        if not M_date == final_time:
            cont_dir_name = str(num_div+1)
            events.append(OrderedDict(
                [('number', num_div+1),
                 ('latitude', -12345),
                 ('longitude', -12345),
                 ('depth', -12345),
                 ('datetime', m_date + num_div*input_dics['interval']),
                 ('magnitude', -12345),
                 ('magnitude_type', 'NAN'),
                 ('author', 'NAN'),
                 ('event_id', 'continuous' + cont_dir_name),
                 ('origin_id', -12345),
                 ('focal_mechanism', False),
                 ('source_duration', False),
                 ('flynn_region', 'NAN'),
                 ('t1', m_date + num_div*input_dics['interval'] -
                  input_dics['preset']),
                 ('t2', M_date + input_dics['offset']),
                 ]))
    else:
        events.append(OrderedDict(
            [('number', 1),
             ('latitude', -12345),
             ('longitude', -12345),
             ('depth', -12345),
             ('datetime', m_date),
             ('magnitude', -12345),
             ('magnitude_type', 'NAN'),
             ('author', 'NAN'),
             ('event_id', 'continuous1'),
             ('origin_id', -12345),
             ('focal_mechanism', False),
             ('source_duration', False),
             ('flynn_region', 'NAN'),
             ('t1', m_date - input_dics['preset']),
             ('t2', M_date + input_dics['offset']),
             ]))
    print('DONE')
    events_QML = Catalog(events=[])
    return events, events_QML

# ##################### neic_catalog_urllib ############################


def neic_catalog_urllib(t_start, t_end, min_latitude,
                        max_latitude, min_longitude,
                        max_longitude, latitude, longitude,
                        radius_min, radius_max, d_min, d_max, mag_min, mag_max,
                        link_neic="http://earthquake.usgs.gov/fdsnws/event/1/query.quakeml?"):
    """
    Function for downloading data from NEIC
    :param t_start:
    :param t_end:
    :param min_latitude:
    :param max_latitude:
    :param min_longitude:
    :param max_longitude:
    :param latitude:
    :param longitude:
    :param radius_min:
    :param radius_max:
    :param d_min:
    :param d_max:
    :param mag_min:
    :param mag_max:
    :param link_neic:
    :return:
    """
    tic = time.perf_counter()

    dir_name = '%s_temp_xml_files' % int(UTCDateTime.now().timestamp)
    os.mkdir(dir_name)

    getVars = {'minmagnitude': str(mag_min),
               'maxmagnitude': str(mag_max),
               'mindepth': str(d_min),
               'maxdepth': str(d_max),
               }

    if None in [latitude, longitude, radius_min, radius_max]:
        if not None in [min_latitude, max_latitude, min_longitude, max_longitude]:
            getVars['minlongitude'] = str(min_longitude)
            getVars['maxlongitude'] = str(max_longitude)
            getVars['minlatitude'] = str(min_latitude)
            getVars['maxlatitude'] = str(max_latitude)
    else:
        getVars['latitude'] = str(latitude)
        getVars['longitude'] = str(longitude)
        getVars['maxradiuskm'] = str(float(radius_max)*111.194)

    getVars['includeallorigins'] = 'true'
    getVars['includeallmagnitudes'] = 'true'
    getVars['producttype'] = 'moment-tensor'

    m_date = UTCDateTime(t_start)
    M_date = UTCDateTime(t_end)
    dur_event = M_date - m_date
    interval = 30.*24.*60.*60.

    num_div = int(dur_event/interval)
    print('#Divisions: %s' % num_div)
    remotefile = False
    if not num_div < 1:
        for i in range(1, num_div+1):
            try:
                print(i, end=',')
                sys.stdout.flush()
                t_start_split = m_date + (i-1)*interval
                t_end_split = m_date + i*interval
                getVars['starttime'] = str(t_start_split)
                getVars['endtime'] = str(t_end_split)

                url_values = urlencodeparse(getVars)
                remotefile = link_neic + url_values
                page = urlopen(remotefile)
                page_content = page.read()
                page_content = page_content.decode("utf-8")

                if 'quakeml' in page_content:
                    with open(os.path.join(dir_name,
                                           'temp_neic_xml_%05i.xml' % i), 'w') \
                            as fid:
                        fid.write(page_content)
                    fid.close()
                else:
                    continue
                page.close()
            except Exception as error:
                print("\nWARNING: %s -- %s\n" % (error, remotefile))
    elif num_div == 0:
        try:
            t_start_split = m_date
            t_end_split = M_date
            getVars['starttime'] = str(t_start_split)
            getVars['endtime'] = str(t_end_split)

            url_values = urlencodeparse(getVars)
            remotefile = link_neic + url_values
            page = urlopen(remotefile)
            page_content = page.read()
            page_content = page_content.decode("utf-8")

            if 'quakeml' in page_content:
                with open(os.path.join(dir_name,
                                       'temp_neic_xml_%05i.xml' % 0), 'w') \
                        as fid:
                    fid.write(page_content)
                fid.close()
            page.close()
        except Exception as error:
            print("\nWARNING: %s -- %s\n" % (error, remotefile))

    try:
        final_time = m_date + num_div*interval
        if (not M_date == final_time) and (not int(dur_event/interval) == 0):
            t_start_split = final_time
            t_end_split = M_date
            getVars['starttime'] = str(t_start_split)
            getVars['endtime'] = str(t_end_split)

            url_values = urlencodeparse(getVars)
            remotefile = link_neic + url_values
            page = urlopen(remotefile)
            page_content = page.read()
            page_content = page_content.decode("utf-8")

            if 'quakeml' in page_content:
                with open(os.path.join(dir_name,
                                       'temp_neic_xml_%05i.xml' % (num_div+1)),
                          'w') as fid:
                    fid.write(page_content)
                fid.close()
            page.close()
    except Exception as error:
        print("\nWARNING: %s\n" % error)

    xml_add = glob.glob(os.path.join(dir_name, 'temp_neic_xml_*.xml'))
    xml_add.sort()
    cat = Catalog()
    print('\nAssembling %s xml files...' % len(xml_add))
    counter = 1
    for x_add in xml_add:
        print(counter, end=',')
        sys.stdout.flush()
        counter += 1
        try:
            cat.extend(readEvents(x_add, format='QuakeML'))
            os.remove(x_add)
        except Exception as error:
            print('[WARNING] %s' % error)
            os.remove(x_add)

    print("\ncleaning up the temporary folder.")
    os.rmdir(dir_name)
    toc = time.perf_counter()
    print('\n%s sec to retrieve the event info form NEIC.' % (toc-tic))
    return cat

# ##################### gcmt_catalog ############################


def gcmt_catalog(t_start, t_end, min_latitude, max_latitude, min_longitude,
                 max_longitude, latitude, longitude, radius_min, radius_max,
                 d_min, d_max, mag_min, mag_max,
                 link_gcmt='http://www.ldeo.columbia.edu/~gcmt/projects/CMT/catalog'):
    """
    Function for downloading data from GCMT
    :param t_start:
    :param t_end:
    :param min_latitude:
    :param max_latitude:
    :param min_longitude:
    :param max_longitude:
    :param latitude:
    :param longitude:
    :param radius_min:
    :param radius_max:
    :param d_min:
    :param d_max:
    :param mag_min:
    :param mag_max:
    :param link_gcmt:
    :return:
    """
    # for the time record
    tic = datetime.now()

    try:
        import obspyDMT
        dmt_path = obspyDMT.__path__[0]
    except Exception as error:
        print("WARNING: %s" % error)
        dmt_path = '.'
    gcmt_cat_path = os.path.join(dmt_path, 'gcmt_catalog')
    if not os.path.exists(gcmt_cat_path):
        os.mkdir(gcmt_cat_path)
        os.mkdir(os.path.join(gcmt_cat_path, 'NEW_MONTHLY'))
        os.mkdir(os.path.join(gcmt_cat_path, 'COMBO'))

    # creating a time list
    t_list = []
    delta_t = int(int(UTCDateTime(t_end)-UTCDateTime(t_start)+1)/86400)

    yymm = []
    for i in range(delta_t + 1):
        t_list.append((UTCDateTime(t_start)+i*60*60*24).strftime('%Y/%m/%d'))
        yy_tmp, mm_tmp, dd_tmp = t_list[i].split('/')
        yymm.append('%s%s' % (yy_tmp, mm_tmp))
    yymmset = set(yymm)
    yymmls = list(yymmset)
    yymmls.sort()

    # starting to search for all events in the time window given by the user:
    cat = Catalog()
    yy_ret = []
    mm_ret = []
    remotefile_add = False

    for i in range(len(yymmls)):
        try:
            yy = yymmls[i][0:4]
            mm = yymmls[i][4:6]
            if int(yy) < 2006:
                month_year = ['jan', 'feb', 'mar', 'apr', 'may', 'june',
                              'july', 'aug', 'sept', 'oct', 'nov', 'dec']
            else:
                month_year = ['jan', 'feb', 'mar', 'apr', 'may', 'jun',
                              'jul', 'aug', 'sep', 'oct', 'nov', 'dec']
            if int(yy) >= 2005:
                new_monthly = 'NEW_MONTHLY'
                file_to_open = os.path.join(gcmt_cat_path, new_monthly,
                                            '%s%s.ndk'
                                            % (month_year[int(mm)-1], yy[-2:]))
                remotefile_add = '%s/%s/%s/%s%s.ndk' \
                                 % (link_gcmt, new_monthly, yy,
                                    month_year[int(mm)-1], yy[-2:])
            else:
                new_monthly = 'COMBO'
                if yy in yy_ret:
                    continue
                file_to_open = os.path.join(gcmt_cat_path, new_monthly,
                                            '%s.qml' % yy)
            if not os.path.exists(file_to_open) and not new_monthly == 'COMBO':
                print('Reading the data from GCMT webpage: %s' % yymmls[i])
                if not os.path.isdir(os.path.join(gcmt_cat_path, new_monthly)):
                    print('Create: %s' % os.path.join(gcmt_cat_path, new_monthly))
                    os.makedirs(os.path.join(gcmt_cat_path, new_monthly))
                remotefile = urlopen(remotefile_add)
                remotefile_read = remotefile.readlines()
                search_fio = open(file_to_open, 'wb')
                search_fio.writelines(remotefile_read)
                search_fio.close()
            if (not os.path.exists(file_to_open)) and (new_monthly == 'COMBO'):
                if not os.path.isdir(os.path.join(gcmt_cat_path, new_monthly)):
                    print('Create: %s' % os.path.join(gcmt_cat_path, new_monthly))
                    os.makedirs(os.path.join(gcmt_cat_path, new_monthly))
                print('WARNING: The following file does not exist: %s' % file_to_open)
                print('WARNING: Reading data from GCMT webpage: %s' % ('http://www.ldeo.columbia.edu/~gcmt/projects/CMT/catalog/COMBO/combo.ndk'))
                print('WARNING: This will take a very long time, refer to:')
                print('https://github.com/kasra-hosseini/obspyDMT/tree/master/obspyDMT/gcmt_catalog/COMBO')
                cat = readEvents('http://www.ldeo.columbia.edu/~gcmt/projects/CMT/catalog/COMBO/combo.ndk')
            else:
                print('Reading the data from local gcmt_catalog: %s' % yymmls[i])
                cat.extend(readEvents(file_to_open))
            yy_ret.append(yy)
            mm_ret.append(mm)
        except Exception as error:
            print("ERROR: %s" % error)

    print('Done reading the data from GCMT webpage.')
    toc = datetime.now()
    print('%s sec to retrieve the event info form GCMT.' % (toc-tic))

    filt1 = 'time >= %s' % t_start
    filt2 = 'time <= %s' % t_end
    cat = cat.filter(filt1, filt2)

    filt1 = 'magnitude >= %s' % mag_min
    filt2 = 'magnitude <= %s' % mag_max
    cat = cat.filter(filt1, filt2)

    filt1 = 'depth >= %s' % (float(d_min)*1000.)
    filt2 = 'depth <= %s' % (float(d_max)*1000.)
    cat = cat.filter(filt1, filt2)

    if None not in [min_latitude, max_latitude, min_longitude, max_longitude]:
        filt1 = 'latitude >= %s' % min_latitude
        filt2 = 'latitude <= %s' % max_latitude
        cat = cat.filter(filt1, filt2)

        filt1 = 'longitude >= %s' % min_longitude
        filt2 = 'longitude <= %s' % max_longitude
        cat = cat.filter(filt1, filt2)

    # final filtering for the remaining requests
    if None not in [latitude, longitude, radius_min, radius_max]:
        index_rm = []
        for i in range(len(cat)):
            e_lat = cat.events[i].preferred_origin().latitude or \
                    cat.events[i].origins[0].latitude
            e_lon = cat.events[i].preferred_origin().longitude or \
                    cat.events[i].origins[0].longitude
            dist = locations2degrees(latitude, longitude, e_lat, e_lon)
            if not radius_min <= dist <= radius_max:
                index_rm.append(i)
        index_rm.sort()
        index_rm.reverse()
        for i in range(len(index_rm)):
            del cat[index_rm[i]]

    return cat

# ##################### isc_catalog ####################################


def isc_catalog(bot_lat=-90, top_lat=90,
                left_lon=-180, right_lon=180,
                ctr_lat=0, ctr_lon=0, radius=180,
                start_time=UTCDateTime() - 30*24*3600,
                end_time=UTCDateTime(),
                min_dep=-10, max_dep=1000, min_mag=0, max_mag=10,
                mag_type='MW', req_mag_agcy='Any',
                rev_comp='REVIEWED'):

    search_domain = 'rectangular'
    if None in [ctr_lat, ctr_lon, radius]:
        if None in [bot_lat, top_lat, left_lon, right_lon]:
            left_lon = '-180'
            right_lon = '180'
            bot_lat = '-90'
            top_lat = '90'
            search_domain = 'rectangular'
    else:
        ctr_lat = str(ctr_lat)
        ctr_lon = str(ctr_lon)
        radius = str(radius)
        search_domain = 'circular'

    if not mag_type:
        mag_type = 'MW'
    else:
        mag_type = mag_type.upper()

    if req_mag_agcy.lower() == 'any':
        req_mag_agcy = ''

    start_time = UTCDateTime(start_time)
    end_time = UTCDateTime(end_time)

    base_url = isc_url_builder(search_domain=search_domain,
                               bot_lat=bot_lat, top_lat=top_lat,
                               left_lon=left_lon, right_lon=right_lon,
                               ctr_lat=ctr_lat, ctr_lon=ctr_lon,
                               radius=radius,
                               start_time=start_time,
                               end_time=end_time,
                               min_dep=min_dep, max_dep=max_dep,
                               min_mag=min_mag, max_mag=max_mag,
                               mag_type=mag_type,
                               req_mag_agcy=req_mag_agcy,
                               rev_comp=rev_comp)

    print("\nURL:\n%s\n" % base_url)

    try_url = 1
    isc_events = Catalog(events=[]) 
    while try_url < 6:
        print("---> Send event request (Try: %s)" % try_url)
        try:
            isc_req = urlopen(base_url)
            isc_contents = isc_req.read()
            isc_events = readEvents(isc_contents)
            try_url = 100
        except Exception as e:
            print("requested content from ISC:\n%s" % e)
        try_url += 1

    remove_index = []
    for i in range(len(isc_events)):
        found_mag_type = False
        for j in range(len(isc_events.events[i].magnitudes)):
            if mag_type in \
                    isc_events.events[i].magnitudes[j].magnitude_type.upper():
                isc_events.events[i].preferred_mag = \
                    isc_events.events[i].magnitudes[j].mag
                isc_events.events[i].preferred_mag_type = \
                    isc_events.events[i].magnitudes[j].magnitude_type
                isc_events.events[i].preferred_author = \
                    isc_events.events[i].magnitudes[j].creation_info.author
                if (min_mag <= isc_events.events[i].preferred_mag <= max_mag):
                    found_mag_type = True
                break
        if not found_mag_type:
            remove_index.append(i)

    if len(remove_index) > 0:
        remove_index.sort(reverse=True)
        for ri in remove_index:
            del isc_events.events[ri]

    return isc_events

# ##################### isc_url_builder ####################################


def isc_url_builder(search_domain='rectangular', bot_lat=-90, top_lat=90,
                    left_lon=-180, right_lon=180,
                    ctr_lat=0, ctr_lon=0, radius=180,
                    start_time=UTCDateTime() - 30*24*3600,
                    end_time=UTCDateTime(),
                    min_dep=-10, max_dep=1000, min_mag=0, max_mag=10,
                    mag_type='MW', req_mag_agcy='Any',
                    rev_comp='reviewed'):
    """
    URL builder for ISC event catalog
    :param search_domain:
    :param bot_lat:
    :param top_lat:
    :param left_lon:
    :param right_lon:
    :param ctr_lat:
    :param ctr_lon:
    :param radius:
    :param start_time:
    :param end_time:
    :param min_dep:
    :param max_dep:
    :param min_mag:
    :param max_mag:
    :param mag_type:
    :param req_mag_agcy:
    :return:
    """
    base_url = 'http://www.isc.ac.uk/cgi-bin/web-db-v4?'
    base_url += 'request=%s' % rev_comp
    base_url += '&out_format=QuakeML'
    if search_domain == 'rectangular':
        base_url += '&searchshape=RECT'
        base_url += '&bot_lat=%s' % bot_lat
        base_url += '&top_lat=%s' % top_lat
        base_url += '&left_lon=%s' % left_lon
        base_url += '&right_lon=%s' % right_lon
        base_url += '&ctr_lat=&ctr_lon=&radius=&max_dist_units=deg'
        base_url += '&srn=&grn='
    elif search_domain == 'circular':
        base_url += '&searchshape=CIRC'
        base_url += '&ctr_lat=%s' % ctr_lat
        base_url += '&ctr_lon=%s' % ctr_lon
        base_url += '&radius=%s' % radius
        base_url += '&max_dist_units=deg'
        base_url += '&bot_lat=&top_lat=&left_lon=&right_lon='
        base_url += '&srn=&grn='
    base_url += '&start_year=%s' % start_time.year
    base_url += '&start_month=%s' % start_time.month
    base_url += '&start_day=%s' % start_time.day
    base_url += '&start_time=%02i%%3A%02i%%3A%02i' % (start_time.hour,
                                              start_time.minute,
                                              start_time.second)
    base_url += '&end_year=%s' % end_time.year
    base_url += '&end_month=%s' % end_time.month
    base_url += '&end_day=%s' % end_time.day
    base_url += '&end_time=%02i%%3A%02i%%3A%02i' % (end_time.hour,
                                            end_time.minute,
                                            end_time.second)
    base_url += '&min_dep=%s' % min_dep
    base_url += '&max_dep=%s' % max_dep
    base_url += '&min_mag=%s' % min_mag
    base_url += '&max_mag=%s' % max_mag
    base_url += '&req_mag_type=%s' % mag_type
    base_url += '&req_mag_agcy=%s' % req_mag_agcy
    
    base_url += '&min_def='
    base_url += '&max_def='
    base_url += '&include_links=on'
    base_url += '&include_magnitudes=on'
    base_url += '&include_headers=on'
    base_url += '&include_comments=on'
    return base_url

# ##################### output_shell_event ####################################


def output_shell_event(events, request):
    """
    output event information to the shell
    :param events:
    :param request:
    :return:
    """
    # output shell
    spaces, events2, header = event_spaces(events=events, request=request)
    header_template = ['{:<'+str(he+2)+'}' for he in spaces]
    row_format = '{}'.format(''.join(header_template))

    if len(events) == 0:
        pass

    elif len(events) <= 50:
        print('\n' + row_format.format(*header))
        print(80 * '-')
        for i in range(len(events2)):
            print((row_format.format(*events2[i].values())).rstrip())
        print(80 * '-' + '\n')

    else:
        print('\n' + row_format.format(*header))
        print(80 * '-')
        print((row_format.format(*events2[0].values())).rstrip())
        print((row_format.format(*events2[1].values())).rstrip())
        print((row_format.format(*events2[2].values())).rstrip())
        print('..')
        print((row_format.format(*events2[-3].values())).rstrip())
        print((row_format.format(*events2[-2].values())).rstrip())
        print((row_format.format(*events2[-1].values())).rstrip())
        print(80 * '-' + '\n')
    return events2, row_format, header

# ##################### event_spaces ############################


def event_spaces(events, request):
    """
    calculate spaces used for table (shell ouput)
    :param events:
    :param request:
    :return:
    """
    header = ['None']
    events2 = copy.deepcopy(events)
    for i in range(len(events2)):
        for item_ev in ['t1', 't2', 'origin_id', 'magnitude_type',
                        'focal_mechanism', 'source_duration']:
            try:
                del events2[i][item_ev]
            except Exception as error:
                print('WARNING: %s' % error)
                pass

        try:
            events2[i]['datetime'] = str(events2[i]['datetime'])[:-8]
        except Exception as error:
            print('WARNING: %s' % error)
            pass

        try:
            if request == 'continuous':
                del events2[i]['flynn_region']
                header = ['#N', 'LAT', 'LON', 'DEP', 'DATETIME', 'MAG',
                          'AUTH', 'EV_ID']
                events2[i]['latitude'] = \
                    "{:>6}".format(int(events2[i]['latitude']))
                events2[i]['longitude'] = \
                    "{:>6}".format(int(events2[i]['longitude']))
                events2[i]['depth'] = \
                    "{:>6}".format(int(events2[i]['depth']))
            else:
                header = ['#N', 'LAT', 'LON', 'DEP', 'DATETIME', 'MAG',
                          'AUTH', 'EV_ID', 'FLY']
                events2[i]['latitude'] = \
                    "{:>8.3f}".format(float(events2[i]['latitude']))
                events2[i]['longitude'] = \
                    "{:>8.3f}".format(float(events2[i]['longitude']))
                events2[i]['depth'] = \
                    int(round(float(events2[i]['depth'])))
        except Exception as error:
            print('WARNING: %s' % error)
            pass

    try:
        k, spaces = [], []
        for i in range(len(events2)):
            # k.append(events2[i].values())
            k.append(list(events2[i].values()))

        for j in range(len(k[0])):
            spaces.append(max([len(str(k[i][j])) for i in range(len(k))]))

        return spaces, events2, header
    except Exception as error:
        print('WARNING: %s' % error)
        pass

# ##################### write_cat_logger ############################


def write_cat_logger(input_dics, eventpath, events, catalog,
                     events2, row_format, header):
    """
    writing outputs from get_time_window
    :param input_dics:
    :param eventpath:
    :param events:
    :param catalog:
    :param events2:
    :param row_format:
    :param header:
    :return:
    """
    if not os.path.isdir(os.path.join(eventpath, 'EVENTS-INFO')):
        os.makedirs(os.path.join(eventpath, 'EVENTS-INFO'))
    input_logger(argus=sys.argv,
                 address=os.path.join(eventpath,
                                      'EVENTS-INFO',
                                      'logger_command.txt'),
                 inputs=input_dics)

    # output catalogue creation info as ASCII
    event_cat = \
        open(os.path.join(eventpath, 'EVENTS-INFO', 'catalog_info.txt'), 'at+')
    st_argus = 'Command line:\n-------------\n'
    for item in sys.argv:
        st_argus += item + ' '
    st_argus += '\n'
    event_cat.writelines(st_argus)
    event_cat.writelines('\n')
    event_cat.writelines('Information about the requested Events:' + '\n\n')
    event_cat.writelines('Number of Events: %s\n' % len(events))
    event_cat.writelines('min datetime: %s\n' % input_dics['min_date'])
    event_cat.writelines('max datetime: %s\n' % input_dics['max_date'])
    event_cat.writelines('min magnitude: %s\n' % input_dics['min_mag'])
    event_cat.writelines('max magnitude: %s\n' % input_dics['max_mag'])
    event_cat.writelines('min latitude: %s\n' % input_dics['evlatmin'])
    event_cat.writelines('max latitude: %s\n' % input_dics['evlatmax'])
    event_cat.writelines('min longitude: %s\n' % input_dics['evlonmin'])
    event_cat.writelines('max longitude: %s\n' % input_dics['evlonmax'])
    event_cat.writelines('min depth: %s\n' % input_dics['min_depth'])
    event_cat.writelines('max depth: %s\n' % input_dics['max_depth'])
    event_cat.writelines('\n\n=====================================\n')
    event_cat.close()

    # output catalogue as ASCII
    event_cat = \
        open(os.path.join(eventpath, 'EVENTS-INFO', 'catalog.txt'), 'at+')
    st_argus = 'Command line:\n-------------\n'
    for item in sys.argv:
        st_argus += item + ' '
    st_argus += '\n'
    event_cat.writelines(st_argus)
    event_cat.writelines('\n')
    event_cat.writelines('#number,event_id,datetime,latitude,longitude,'
                         'depth,magnitude,magnitude_type,author,'
                         'flynn_region,mrr,mtt,mpp,mrt,mrp,mtp,'
                         'stf_func,stf_duration,t1,t2\n')
    for ev in events:
        if ev['focal_mechanism']:
            mrr = ev['focal_mechanism'][0]
            mtt = ev['focal_mechanism'][1]
            mpp = ev['focal_mechanism'][2]
            mrt = ev['focal_mechanism'][3]
            mrp = ev['focal_mechanism'][4]
            mtp = ev['focal_mechanism'][5]
        else:
            mrr = None
            mtt = None
            mpp = None
            mrt = None
            mrp = None
            mtp = None
        if ev['source_duration']:
            stf_shape = ev['source_duration'][0]
            stf_dur = ev['source_duration'][1]
        else:
            stf_shape = None
            stf_dur = None
        event_cat.writelines('%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,'
                             '%s,%s,%s,%s,%s,%s,%s\n'
                             % (ev['number'], ev['event_id'], ev['datetime'],
                                ev['latitude'], ev['longitude'],
                                ev['depth'], ev['magnitude'],
                                ev['magnitude_type'], ev['author'],
                                ev['flynn_region'],
                                mrr, mtt, mpp, mrt, mrp, mtp,
                                stf_shape, stf_dur,
                                ev['t1'],
                                ev['t2']))
    event_cat.writelines('\n')
    event_cat.close()

    event_file = open(os.path.join(eventpath, 'EVENTS-INFO',
                                   'event_list_pickle'), 'wb')
    pickle.dump(events, event_file, protocol=2)
    event_file.close()

    # output catalogue as ASCII but in table form
    try:
        st_argus = 'Command line:\n-------------\n'
        for item in sys.argv:
            st_argus += item + ' '
        st_argus += '\n'
        event_table = open(os.path.join(eventpath, 'EVENTS-INFO',
                                        'catalog_table.txt'), 'at+')
        event_table.writelines(st_argus)
        event_table.writelines('\n' + row_format.format(*header))
        event_table.writelines('\n' + '-'*80 + '\n')
        for i in range(len(events2)):
            event_table.writelines(
                (row_format.format(*events2[i].values())).rstrip() + '\n')
        event_table.writelines('-'*80 + '\n')
    except Exception as err:
        print('\nCouldn\'t write catalog object to ASCII file as:\n %s\n' \
              'Proceed without ..\n' % err)

    # output catalogue as QUAKEML / JSON files
    try:
        if not input_dics['event_catalog'].lower() == 'local':
            if os.path.isfile(os.path.join(eventpath, 'EVENTS-INFO', 'catalog.ml.pkl')):
                stored_catalog = Catalog(events=[])
                try:
                    stored_catalog = pickle.load(open(os.path.join(eventpath,
                                                                   'EVENTS-INFO',
                                                                   'catalog.ml.pkl'), "rb"))
                except Exception as err:
                    pass
                for stored_ev in stored_catalog:
                    if not stored_ev in catalog:
                        catalog.append(stored_ev)
            pickle.dump(catalog,
                        open(os.path.join(eventpath, 'EVENTS-INFO', 'catalog.ml.pkl'), "wb"))
    except Exception as err:
        print('\nCouldn\'t write catalog object to catalog.ml.pkl as:\n>>:\t %s\n' \
              'Proceed without ..\n' % err)
    try:
        if not input_dics['event_catalog'].lower() == 'local':
            catalog.write(os.path.join(eventpath, 'EVENTS-INFO', 'catalog.zmap'),
                          format="ZMAP")
    except Exception as err:
        print('\nCouldn\'t write catalog object to JSON as:\n>>:\t %s\n' \
              'Proceed without ..\n' % err)

# ##################### sort_catalogue ############################


def sort_catalogue(cat):
    """
    sort catalogue of retrieved events chronological.
    :param cat:
    :return:
    """
    k = [[event, event.origins[0].time] for event in cat]
    k.sort(key=lambda x: x[1], reverse=True)
    events = [event[0] for event in k]
    cat = Catalog(events=events)
    return cat

# ##################### mag_duration ###################################


def mag_duration(mag, type_curve=1):
    """
    calculate the source duration out of magnitude
    type_curve can be 1, 2, 3:
    1: 2005-2014
    2: 1976-1990
    3: 1976-2014
    :param mag:
    :param type_curve:
    :return:
    """
    if type_curve == 1:
        half_duration = 0.00272*np.exp(1.134*mag)
    elif type_curve == 2:
        half_duration = 0.00804*np.exp(1.025*mag)
    elif type_curve == 3:
        half_duration = 0.00392*np.exp(1.101*mag)
    else:
        sys.exit('%s Type for magnitude to source duration conversion is not '
                 'implemented' % type_curve)
    source_duration = round(half_duration, 3)*2
    return ['triangle', source_duration]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -------------------------------------------------------------------
#   Filename:  utility_codes.py
#   Purpose:   collection of helping scripts and utility codes
#   Author:    Kasra Hosseini
#   Email:     kasra.hosseinizad@earth.ox.ac.uk
#   License:   GNU Lesser General Public License, Version 3
# -------------------------------------------------------------------

# -----------------------------------------------------------------------
# ----------------Import required Modules (Python and Obspy)-------------
# -----------------------------------------------------------------------
from datetime import datetime, timedelta
import fnmatch
import glob
import numpy as np
from obspy.core import read
try:
    from obspy.clients.fdsn import URL_MAPPINGS
except:
    from obspy.fdsn.header import URL_MAPPINGS
try:
    from obspy.geodetics import locations2degrees
except:
    from obspy.core.util import locations2degrees
try:
    from obspy.signal.util import next_pow_2 as nextpow2
except:
    from obspy.signal.util import nextpow2

import os
import pickle
import smtplib
import sys
import time

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

# ##################### header_printer ##################################


def header_printer():
    """
    clear the screen and print the welcome message.
    :return:
    """
    os.system('clear')
    print(80*'-')
    print('\t\t   obspyDMT (obspy Data Management Tool)\n')
    print('\tPython Toolbox for Retrieving, Processing and Management of')
    print('\t\t\tLarge Seismological Datasets\n')
    print(':copyright:')
    print('The ObsPy Development Team (devs@obspy.org)\n')
    print('Developed by Kasra Hosseini')
    print('email: kasra.hosseinizad@earth.ox.ac.uk\n')
    print(':license:')
    print('GNU General Public License, Version 3')
    print('(http://www.gnu.org/licenses/gpl-3.0-standalone.html)')
    print(80*'-' + '\n')

# ##################### goodbye_printer ##################################


def goodbye_printer(input_dics, t1_pro):
    """
    print the goodbye message which contains information about duration of
    the run and size of the directory.
    :param input_dics:
    :param t1_pro:
    :return:
    """
    print("\n\n==================================================")
    print("obspyDMT main program has finished!\n")
    try:
        #size = getFolderSize(input_dics['datapath'])
        #size /= (1024.**2)
        #print("Info:")
        #print("* The directory contains %s MB of data." \
        #      % "{:.3f}".format(float(size)))
        print(input_dics['datapath'])
        print("* Total time of execution: %s (h:m:s)" \
              % str(timedelta(seconds=round(float(time.time() - t1_pro)))))
        print("==================================================\n\n")
    except Exception as error:
        print('ERROR: %s' % error)
        pass

# ##################### print_data_sources ##################################


def print_data_sources():
    """
    Function to print available data providers
    :return:
    """
    print("\n--------------------------")
    print("list of all shortcut names")
    print("--------------------------\n")
    for key in sorted(URL_MAPPINGS.keys()):
        print("{0:<15} {1}".format(key,  URL_MAPPINGS[key]))
    print("iris-federator")
    print("eida-routing")
    print("\n============================================================")
    print("This is the list of all shortcut names which can be used for")
    print("--data_source option.")
    print("However, FDSN base URLs can be entered directly as well.")
    print("============================================================")
    sys.exit()

# ##################### print_event_catalogs ##################################


def print_event_catalogs():
    """
    Function to print available event catalogs
    :return:
    """
    print("\n------------------------")
    print("supported event catalogs")
    print("------------------------\n")
    for ev_cat in ['LOCAL', 'NEIC_USGS', 'GCMT_COMBO', 'IRIS', 'NCEDC',
                   'USGS', 'INGV', 'ISC']:
        print(ev_cat)

    print("\n============================================================")
    print("This is the list of all available event catalogs that can be")
    print("used for --event_catalog option.")
    print("============================================================")
    sys.exit()

# ##################### print_syngine_models ##################################


def print_syngine_models():
    """
    Function to print available syngine models
    :return:
    """
    print("\n------------------------")
    print("Available syngine models")
    print("------------------------\n")

    from obspy.clients.syngine import Client as Client_syngine
    client_syngine = Client_syngine()
    avail_syngine_models = client_syngine.get_available_models()

    for count_mod, avail_mod in enumerate(avail_syngine_models.keys()):
        print('%s. %s' % (count_mod+1, avail_mod))
    print('\n')
    for count_mod, avail_mod in enumerate(avail_syngine_models.keys()):
        print("------------------------")
        print('%s. %s' % (count_mod+1, avail_mod))
        for inf in avail_syngine_models[avail_mod].keys():
            print("%s: %s" % (inf, avail_syngine_models[avail_mod][inf]))

    print("\n============================================================")
    print("This is the list of all available syngine models that can be")
    print("used for --syngine_bg_model option.")
    print("============================================================")
    sys.exit()

# ##################### create_folders_files ############################


def create_folders_files(event, eventpath, input_dics):
    """
    create required directories and files for one event
    :param event:
    :param eventpath:
    :param input_dics:
    :return:
    """
    try:
        for t_dir in ['raw', 'resp', 'info']:
            tar_dir = os.path.join(eventpath, event['event_id'], t_dir)
            if not os.path.isdir(tar_dir):
                os.makedirs(tar_dir)

        inp_file = open(os.path.join(eventpath, event['event_id'],
                                     'info', 'input_dics.pkl'), 'wb')
        pickle.dump(input_dics, inp_file, protocol=2)
        inp_file.close()

        event_file = open(os.path.join(eventpath, event['event_id'],
		          'info', 'event.pkl'), 'wb')
        pickle.dump(event, event_file, protocol=2)
        event_file.close()

        report = open(os.path.join(eventpath, event['event_id'],
                                   'info', 'report_st'), 'at+')
        report.close()

        exception_file = open(os.path.join(eventpath, event['event_id'],
                                           'info', 'exception'), 'at+')
        exception_file.writelines('\n' + event['event_id'] + '\n')
        exception_file.close()

        syn_file = open(os.path.join(eventpath, event['event_id'],
                                     'info', 'station_event'), 'at+')
        syn_file.close()

    except Exception as error:
        print('ERROR: %s' % error)
        pass

# ##################### read_list_stas ##################################


def read_list_stas(add_list, normal_mode_syn, specfem3D):
    """
    read a list of stations instead of checking the availability
    :param add_list:
    :param normal_mode_syn:
    :param specfem3D:
    :return:
    """
    print('\n---------------------------------------------')
    print('INFO:')
    print('Format of the station list:')
    print('net,sta,loc,cha,lat,lon,ele,depth,data_source')
    print('---------------------------------------------\n\n')

    list_stas_fio = open(add_list, 'rt')
    list_stas = list_stas_fio.readlines()
    for sta in range(len(list_stas)):
        if not list_stas[sta].startswith('\n'):
            list_stas[sta] = [x.strip() for x in list_stas[sta].split(',')]

    final_list = []
    if specfem3D:
        for sta in range(len(list_stas)):
            for chan in ['MXE', 'MXN', 'MXZ']:
                st_id = 'SY_%s_S3_%s' % (list_stas[sta][1], chan)
                final_list.append(['SY', list_stas[sta][1],
                                   'S3', chan,
                                   list_stas[sta][4],
                                   list_stas[sta][5],
                                   list_stas[sta][6],
                                   list_stas[sta][7],
                                   'IRIS', st_id])
    elif normal_mode_syn:
        for sta in range(len(list_stas)):
            for chan in ['LXE', 'LXN', 'LXZ']:
                st_id = 'SY_%s_S1_%s' % (list_stas[sta][1], chan)
                final_list.append(['SY', list_stas[sta][1],
                                   'S1', chan,
                                   list_stas[sta][4],
                                   list_stas[sta][5],
                                   list_stas[sta][6],
                                   list_stas[sta][7],
                                   'IRIS', st_id])
    else:
        for sta in range(len(list_stas)):
            st_id = '%s_%s_%s_%s' % (list_stas[sta][0],
                                     list_stas[sta][1],
                                     list_stas[sta][2],
                                     list_stas[sta][3])
            final_list.append([list_stas[sta][0], list_stas[sta][1],
                               list_stas[sta][2], list_stas[sta][3],
                               list_stas[sta][4], list_stas[sta][5],
                               list_stas[sta][6], list_stas[sta][7],
                               list_stas[sta][8], st_id])
    return final_list

# ##################### read_event_dic ##############################


def read_event_dic(address):
    """
    Reads event dictionary ("info" folder)
    :param address:
    :return:
    """
    if not os.path.isabs(address):
        address = os.path.abspath(address)

    if os.path.basename(address) == 'info':
        target_add = [address]
    elif locate(address, 'info'):
        target_add = locate(address, 'info')
    else:
        print('Error: There is no "info" directory in %s' % address)
        target_add = None

    if len(target_add) > 1:
        print("[ERROR] there are more than one directory to " \
              "read the event info:")
        for tar_add in target_add:
            print(tar_add)
        sys.exit('\nonly one directory should be specified to proceed!')

    event_dic = False
    for t_add in target_add:
        if os.path.isfile(os.path.join(t_add, 'event.pkl')):
            ev_file_fio = open(os.path.join(t_add, 'event.pkl'), 'rb')
            event_dic = pickle.load(ev_file_fio)
        else:
            sys.exit('[ERROR] event.pkl can not be found in: %s' % t_add)
    return event_dic

# ##################### read_station_event ##############################


def read_station_event(address):
    """
    reads the station_event file ("info" folder)
    :param address:
    :return:
    """
    if not os.path.isabs(address):
        address = os.path.abspath(address)

    if os.path.basename(address) == 'info':
        target_add = [address]
    elif locate(address, 'info'):
        target_add = locate(address, 'info')
    else:
        print('[ERROR] There is no "info" directory in:\n%s' % address)
        target_add = None

    sta_ev = []
    for t_add in target_add:
        if os.path.isfile(os.path.join(t_add, 'station_event')):
            sta_file_open = open(os.path.join(t_add, 'station_event'), 'rt')
        else:
            print('=====================================')
            print('station_event could not be found')
            print('start creating the station_event file')
            print('=====================================')
            create_station_event(address=t_add)
            sta_file_open = open(os.path.join(t_add, 'station_event'), 'rt')
        sta_file = sta_file_open.readlines()
        sta_ev_tmp = []
        for s_file in sta_file:
            sta_ev_tmp.append(s_file.split(','))
        sta_ev.append(sta_ev_tmp)
    return sta_ev

# ##################### create_station_event ############################


def create_station_event(address):
    """
    create the station_event file ("info" folder)
    :param address:
    :return:
    """
    event_address = os.path.dirname(address)
    if os.path.isdir(os.path.join(event_address, 'raw')):
        sta_address = os.path.join(event_address, 'raw')
    elif os.path.isdir(os.path.join(event_address, 'processed')):
        sta_address = os.path.join(event_address, 'processed')
    else:
        sys.exit('[ERROR] There is no reference (raw or processed) '
                 'to create station_event file!')

    ls_stas = glob.glob(os.path.join(sta_address, '*'))
    ls_stas.sort()

    print('%s stations found in %s' % (len(ls_stas), sta_address))

    for i in range(len(ls_stas)):
        print(i),
        sta_file_open = open(os.path.join(address, 'station_event'), 'at')
        try:
            sta = read(ls_stas[i])[0]
        except Exception as e:
            print('[WARNING] NOT readable: %s\n%s' % (ls_stas[i], e))
            sta = None
        try:
            sta_stats = sta.stats
            sta_info = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,iris,\n' \
                       % (sta_stats.network,
                          sta_stats.station,
                          sta_stats.location,
                          sta_stats.channel,
                          sta_stats.sac.stla,
                          sta_stats.sac.stlo,
                          sta_stats.sac.stel,
                          sta_stats.sac.stdp,
                          os.path.basename(event_address),
                          sta_stats.sac.evla,
                          sta_stats.sac.evlo,
                          sta_stats.sac.evdp,
                          sta_stats.sac.mag)
        except Exception as error:
            print('\n[WARNING] Can not read all the required information ' \
                  'from the headers, some of them are presumed!')
            print(error)
            sta_info = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,iris,\n' \
                       % (sta_stats.network,
                          sta_stats.station,
                          sta_stats.location,
                          sta_stats.channel,
                          -12345.0, -12345.0, -12345.0, -12345.0,
                          os.path.basename(event_address),
                          -12345.0, -12345.0, -12345.0, -12345.0)
        sta_file_open.writelines(sta_info)
        sta_file_open.close()
    print('station_event file is created in %s' \
          % os.path.join(address, 'station_event'))

# ##################### locate ##########################################


def locate(root='.', target='info', num_matches=-1):
    """
    locates a subdirectory within a directory
    :param root:
    :param target:
    :param num_matches:
    :return:
    """
    matches = []
    for root, dirnames, filenames in os.walk(root):
        for dirname in fnmatch.filter(dirnames, target):
            matches.append(os.path.join(root, dirname))
        if num_matches > 0:
            if len(matches) >= num_matches:
                break

    return matches

# ##################### convert_to_sac ########################################


def convert_to_sac(tr, save_path, sta_ev_arr):
    """
    convert tr format to SAC and try to fill in some header information
    :param tr:
    :param save_path:
    :param sta_ev_arr:
    :return:
    """
    tr.write(save_path, format='SAC')
    tr = read(save_path)[0]

    # start filling in some header information
    try:
        tr.stats.sac.stla = float(sta_ev_arr[4])
    except:
        pass
    try:
        tr.stats.sac.stlo = float(sta_ev_arr[5])
    except:
        pass
    try:
        tr.stats.sac.stel = float(sta_ev_arr[6])
    except:
        pass
    try:
        tr.stats.sac.stdp = float(sta_ev_arr[7])
    except:
        pass
    try:
        tr.stats.sac.evla = float(sta_ev_arr[10])
    except:
        pass
    try:
        tr.stats.sac.evlo = float(sta_ev_arr[11])
    except:
        pass
    try:
        tr.stats.sac.evdp = float(sta_ev_arr[12])
    except:
        pass
    try:
        tr.stats.sac.mag = float(sta_ev_arr[13])
    except:
        pass
    try:
        tr.stats.sac.cmpaz = float(sta_ev_arr[14])
    except:
        pass
    try:
        tr.stats.sac.cmpinc = float(sta_ev_arr[15])
    except:
        pass
    return tr

# ##################### calculate_time_phase ##################################


def calculate_time_phase(event, sta, bg_model='iasp91'):
    """
    calculate arrival time of the requested phase
    :param event:
    :param sta:
    :param bg_model:
    :return:
    """
    phase_list = ['P', 'Pdiff', 'PKIKP']

    time_ph = 0
    ev_lat = event['latitude']
    ev_lon = event['longitude']
    evdp = abs(float(event['depth']))
    sta_lat = float(sta[4])
    sta_lon = float(sta[5])
    dist = locations2degrees(ev_lat, ev_lon, sta_lat, sta_lon)

    try:
        from obspy.taup import tau
        tau_bg = tau.TauPyModel(model=bg_model)
    except:
        tau_bg = False

    if not tau_bg:
        try:
            try:
                from obspy.taup import getTravelTimes
            except:
                from obspy.taup.taup import getTravelTimes
            tt = getTravelTimes(dist, evdp)
            flag = False
            for ph in phase_list:
                for i in range(len(tt)):
                    if tt[i]['phase_name'] == ph:
                        flag = True
                        time_ph = tt[i]['time']
                        break
                    else:
                        continue
            if not flag:
                time_ph = 0
        except:
            time_ph = 0
    else:
        for ph in phase_list:
            try:
                tt = tau_bg.get_travel_times(evdp, dist,
                                             phase_list=[ph])[0].time
                if not tt:
                    time_ph = 0
                    continue
                else:
                    time_ph = tt
                    break
            except:
                time_ph = 0

    t_start = event['t1'] + time_ph
    t_end = event['t2'] + time_ph

    return t_start, t_end

# ##################### plot_filter_station ###############################


def plot_filter_station(input_dics, sta_ev):
    """
    check whether the station can pass the criteria
    :param input_dics:
    :param sta_ev:
    :return:
    """
    if not fnmatch.fnmatch(sta_ev[0], input_dics['net']):
        return False
    if not fnmatch.fnmatch(sta_ev[1], input_dics['sta']):
        return False
    if not fnmatch.fnmatch(sta_ev[2], input_dics['loc']):
        return False
    if not fnmatch.fnmatch(sta_ev[3], input_dics['cha']):
        return False
    if isinstance(input_dics['mlat_rbb'], float):
        if not float(sta_ev[4]) <= input_dics['Mlat_rbb']:
            return False
        if not float(sta_ev[4]) >= input_dics['mlat_rbb']:
            return False
        if not float(sta_ev[5]) <= input_dics['Mlon_rbb']:
            return False
        if not float(sta_ev[5]) >= input_dics['mlon_rbb']:
            return False
    return True

# ##################### getFolderSize ###################################


def getFolderSize(folder):
    """
    returns the size of a folder in bytes
    :param folder:
    :return:
    """
    total_size = os.path.getsize(folder)
    for item in os.listdir(folder):
        itempath = os.path.join(folder, item)
        if os.path.isfile(itempath):
            total_size += os.path.getsize(itempath)
        elif os.path.isdir(itempath):
            total_size += getFolderSize(itempath)
    return total_size

# ##################### send_email ######################################


def send_email(input_dics):
    """
    send email to the specified "email" address
    :param input_dics:
    :return:
    """
    t2_str = datetime.now()

    fromaddr = 'obspyDMT'
    toaddrs = input_dics['email']
    msg = "request finished at:\n%s" % t2_str
    try:
        server = smtplib.SMTP('localhost')
        server.sendmail(fromaddr, toaddrs, msg)
        print('\n*********************************************')
        print('Sending email to the following email-address:')
        print(input_dics['email'])
        print('*********************************************')
    except Exception as error:
        print('\n**************************')
        print('Email was not sent, ERROR:')
        print(error)
        print('**************************')
        # err_info = traceback.extract_tb(sys.exc_info()[2])

# ##################### check_par_jobs ######################################


def check_par_jobs(jobs, sleep_time=1):
    """
    check whether all the parallel jobs are finished or not
    :param jobs:
    :param sleep_time:
    :return:
    """
    pp_flag = True
    while pp_flag:
        for proc in jobs:
            if proc.is_alive():
                time.sleep(sleep_time)
                pp_flag = True
                break
            else:
                pp_flag = False
    if not pp_flag:
        print('\n\n================================')
        print('All %s processes are finished...' % len(jobs))
        print('================================')

# ##################### spectrum_calc ##################################


def spectrum_calc(tr):
    """
    Simple code to calculate the spectrum of a trace
    :param tr: obspy Trace
    :return:
    freqs, spec_tr
    which are frequencies and spectrum of the trace

    To plot the spectrum:

    import matplotlib.pyplot as plt
    plt.loglog(freqs, spec_tr)
    """

    nfft = nextpow2(tr.stats.npts)
    freqs = np.fft.rfftfreq(nfft) / tr.stats.delta

    spec_tr = np.abs(np.fft.rfft(tr.data, n=nfft))

    return freqs, spec_tr

# ------------------ geocen_calc ---------------------------


def geocen_calc(geog_lat):
    """
    Calculate geocentric latitudes
    :param geog_lat:
    :return:
    """
    fac = 0.993305621334896

    colat = 90.0 - geog_lat
    if abs(colat) < 1.0e-5:
        colat = np.sign(colat)*1.0e-5
    # arg = colat*rpd
    colat *= np.pi/180.
    colat_sin = np.sin(colat)
    if colat_sin < 1.0e-30:
        colat_sin = 1.0e-30
    # geocen=pi2-atan(fac*cos(arg)/(max(1.0e-30,sin(arg))))
    geocen_colat = np.pi/2. - np.arctan(fac*np.cos(colat)/colat_sin)
    geocen_colat = geocen_colat*180./np.pi
    geocen_lat = 90.0 - geocen_colat

    return geocen_lat

# ----------------------------------------------------------------------------
# process_unit:
# if len(st) > 1:
#     st.merge(method=1, fill_value=0, interpolation_samples=0)
#     gap_fio = open(os.path.join(target_path, 'info',
#                                 'waveform_gap.txt'), 'a+')
#     gap_msg = '%s.%s.%s.%s\t%s\n' % (st[0].stats.network,
#                                      st[0].stats.station,
#                                      st[0].stats.location,
#                                      st[0].stats.channel,
#                                      'instrument_correction')
#     gap_fio.writelines(gap_msg)
#     gap_fio.close()



import numpy as np
from scipy.spatial import cKDTree

# ------------------ SphericalNearestNeighbour ---------------------------


class SphericalNearestNeighbour():
    """
    Spherical nearest neighbour queries using scipy's fast kd-tree
    implementation.
    """
    def __init__(self, lat, lon, el_dp, eradius=6371009):
        cart_data = self.spherical2cartesian(lat, lon, el_dp, eradius)
        self.kd_tree = cKDTree(data=cart_data, leafsize=10)
        self.eradius = eradius

    def query(self, lat, lon, el_dp, k=1):
        points = self.spherical2cartesian(lat, lon, el_dp, self.eradius)
        d, i = self.kd_tree.query(points, k=k)
        return d, i

    def query_pairs(self, maximum_distance):
        return self.kd_tree.query_pairs(maximum_distance)

    def spherical2cartesian(self, lat, lon, el_dp, eradius):
        """
        Converts a list of :class:`~obspy.fdsn.download_status.Station`
        objects to an array of shape(len(list), 3) containing x/y/z in meters.
        """
        shape = len(lat)
        r = eradius + el_dp
        # Convert data from lat/lng to x/y/z.
        colat = 90.0 - lat
        cart_data = np.empty((shape, 3), dtype=np.float64)
        cart_data[:, 0] = r * np.sin(np.deg2rad(colat)) * \
        np.cos(np.deg2rad(lon))
        cart_data[:, 1] = r * np.sin(np.deg2rad(colat)) * \
        np.sin(np.deg2rad(lon))
        cart_data[:, 2] = r * np.cos(np.deg2rad(colat))
        return cart_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name="obspyDMT",
    version="2.2.10",
    keywords=["obspyDMT", "obspy", "seismology", "geophysics"],
    description="obspyDMT: A Python Toolbox for Retrieving, Processing and "
                 "Management of Seismological Datasets",
    long_description="""\
obspyDMT (obspy Data Management Tool) is a command line tool for retrieving,
processing and management of seismological datasets in a fully automatic way.
""",
    author=u"Kasra Hosseini",
    author_email="k.hosseinizad@gmail.com",
    url="https://github.com/kasra-hosseini/obspyDMT",
    license="GNU Lesser General Public License, Version 3",
    platforms="OS Independent",
    packages=["obspyDMT", "obspyDMT.utils"],
    # this is needed for "pip install instaseis==dev"
    download_url="https://github.com/kasra-hosseini/obspyDMT/archive/master.zip",
    install_requires=[
        "obspy>=1.2.0,<2.0.0",
        "matplotlib>=3.2.0",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics"],
    entry_points={
        'console_scripts': [
            'obspyDMT = obspyDMT.obspyDMT:main',
        ],
    }
)



# ------------- required Python and obspy modules are imported in this part
from obspy.core import read
import os
import subprocess
import sys
import errno

from .utils.utility_codes import convert_to_sac
# -----------------------------------------------------------------------------

# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# ===================== YOU CAN CHANGE THE FOLLOWING FUNCTION =================
# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# * IMPORTANT *
# The following function (process_unit) is in the waveform level.
# This means that you can write your process unit for one trace, and
# obspyDMT uses this function to pre-process all your waveforms,
# either right after retrieval or as a separate step:
# obspyDMT --datapath /your/dataset --local

# ========== process_unit has the following arguments:
# Use the following parameters to write your process_unit:
# 1. tr_add: address of one trace in your dataset. You can use that to
# read in the data.
# 2. target_path: address of the event that should be processed.
# 3. input_dics: dictionary that contains all the inputs.
# 4. staev_ar: an array that contains the following information:
# net, sta, loc, cha, station latitude, station longitude, station elevation,
# station depth


def process_unit(tr_add, target_path, input_dics, staev_ar):
    """
    processing unit, adjustable by the user
    :param tr_add: address of one trace in your dataset. You can use that to
    read in the data.
    :param target_path: address of the event that should be processed.
    :param input_dics: dictionary that contains all the inputs.
    :param staev_ar: an array that contains the following information:
           net, sta, loc, cha, station latitude, station longitude,
           station elevation, station depth
    :return:
    """
    # -------------- read the waveform, deal with gaps ------------------------
    # 1. read the waveform and create an obspy Stream object
    try:
        st = read(tr_add)
    except Exception as error:
        print('WARNING: %s' % error)
        return False
    # 2. in case that there are more than one waveform in a Stream (this can
    # happen due to some gaps in the waveforms) merge them.
    if len(st) > 1:
        try:
            st.merge(method=1, fill_value=0, interpolation_samples=0)
        except Exception as error:
            print('WARNING: %s' % error)
            return False
    # 3. Now, there is only one waveform, create a Trace
    tr = st[0]

    # -------------- path to save the processed waveform ----------------------
    # Before entering to the actual processing part of the code,
    # we define some paths to be used later:
    # you can adjust it as you want, here is just one example
    
    # If pathlib is installed, one can use: (suggested by ghraecakter)
    # pathlib.Path(os.path.join(target_path, 'processed')).mkdir(exist_ok=True)
    try:
        os.mkdir(os.path.join(target_path, 'processed'))
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    # save_path is the address that will be used to save the processed data
    save_path = os.path.join(target_path, 'processed', tr.id)
    if os.path.isfile(save_path) and (not input_dics['force_process']):
        return False

    # -------------- PROCESSING -----------------------------------------------
    tr = convert_to_sac(tr, save_path, staev_ar)
    tr.write(save_path, format='SAC')

    p = subprocess.Popen(['sac'],
                         stdout=subprocess.PIPE,
                         stdin=subprocess.PIPE,
                         stderr=subprocess.STDOUT)

    s = \
        'read ' + save_path + '\n' + \
        'rmean' + '\n' + \
        'rtrend' + '\n' + \
        'taper' + '\n' + \
        'fft' + '\n' + \
        'keepam' + '\n' + \
        'p1' + '\n' + \
        'loglog' + '\n' + \
        'xlim 1e-5 100' + '\n' + \
        'p1' + '\n' + \
        'saveimg ' + save_path + '.pdf\n' + \
        'quit\n'

    if sys.version_info > (3, 0):
        out = p.communicate(bytes(s, "utf-8"))
    else:
        out = p.communicate(bytes(s))
    print(out[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -------------------------------------------------------------------
#   Filename:  local_handler.py
#   Purpose:   handling local processing/plotting in obspyDMT
#   Author:    Kasra Hosseini
#   Email:     kasra.hosseinizad@earth.ox.ac.uk
#   License:   GNU Lesser General Public License, Version 3
# -------------------------------------------------------------------

# -----------------------------------------------------------------------
# ----------------Import required Modules (Python and Obspy)-------------
# -----------------------------------------------------------------------
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import input as raw_input_built
import matplotlib.pyplot as plt
import multiprocessing
import numpy as np
try:
    from obspy.imaging.beachball import beach as Beach
except:
    from obspy.imaging.beachball import beachball as Beach
from obspy import UTCDateTime, read
try:
    from obspy.geodetics import locations2degrees
except:
    from obspy.core.util import locations2degrees
try:
    from obspy.geodetics.base import gps2dist_azimuth as gps2DistAzimuth
except:
    try:
        from obspy.geodetics import gps2DistAzimuth
    except:
        from obspy.core.util import gps2DistAzimuth
import os
import sys

from .data_handler import update_sta_ev_file
from .kml_handler import create_ev_sta_kml
from .utility_codes import locate, check_par_jobs, plot_filter_station

# ###################### process_data #########################################


def process_data(input_dics, event):
    """
    prepare the target directories and pass them to process_serial_parallel
    :param input_dics:
    :param event:
    :return:
    """
    target_path = locate(input_dics['datapath'], event['event_id'], num_matches=1)

    if len(target_path) == 0:
        return
    if len(target_path) > 1:
        print("[LOCAL] more than one path was found for one event:")
        print(target_path)
        print("use the first one:")
        target_path = target_path[0]
        print(target_path)
    else:
        print("[LOCAL] Path:")
        target_path = target_path[0]
        print(target_path)
    print("[INFO] update station_event file...")
    update_sta_ev_file(target_path, event)
    sta_ev_arr = np.loadtxt(os.path.join(target_path, 'info', 'station_event'),
                            delimiter=',', dtype=bytes, ndmin=2).astype(str)
    sta_ev_arr = sta_ev_arr.astype(object)
    if input_dics['select_data']:
        sta_ev_arr = select_data(deg_step=float(input_dics['select_data']),
                                 sta_ev=sta_ev_arr)
    if len(sta_ev_arr) > 0:
        if len(np.shape(sta_ev_arr)) == 1:
            sta_ev_arr = np.reshape(sta_ev_arr, [1, len(sta_ev_arr)])
        process_serial_parallel(sta_ev_arr, input_dics, target_path)
    else:
        print("[LOCAL] no waveform to process for %s!" % target_path)

# ###################### select_data #################################


def select_data(deg_step, sta_ev):
    """
    select stations every deg_step
    :param deg_step:
    :param sta_ev:
    :return:
    """
    lat_step = deg_step
    lon_step = deg_step
    eradius = 6371
    lats = np.arange(-90, 90+lat_step, lat_step)
    lons = np.arange(-180, 180+lon_step, lon_step)
    grd_points = np.empty([len(lats)*len(lons), 2])

    counter = 0
    for lat in lats:
        for lon in lons:
            grd_points[counter, 0] = lat
            grd_points[counter, 1] = lon
            counter += 1

    from .utils.spherical_nearest import SphericalNearestNeighbour
    kd = SphericalNearestNeighbour(grd_points[:, 0],
                                   grd_points[:, 1],
                                   np.zeros(len(grd_points[:, 1])),
                                   eradius=eradius)
    dists, indxs = kd.query(sta_ev[:, 4].astype(np.float),
                            sta_ev[:, 5].astype(np.float),
                            np.zeros(len(sta_ev[:, 4])),
                            1)
    indxs_unique = np.unique(indxs, return_index=True)
    sta_ev_arr_unique = sta_ev[indxs_unique]
    return sta_ev_arr_unique

# ###################### process_serial_parallel ##############################


def process_serial_parallel(sta_ev_arr, input_dics, target_path):
    """
    run the processing unit in parallel or in serial
    :param sta_ev_arr:
    :param input_dics:
    :param target_path:
    :return:
    """
    if input_dics['parallel_process']:
        start = 0
        end = int(len(sta_ev_arr))
        if end < input_dics['process_np']:
            req_proc = end
        else:
            req_proc = input_dics['process_np']
        step = (end - start) / req_proc + 1
        step = int(step)

        jobs = []
        for index in range(req_proc):
            starti = start + index * step
            endi = min(start + (index + 1) * step, end)
            if starti == endi:
                break
            p = multiprocessing.Process(target=process_core_iterate,
                                        args=(sta_ev_arr, input_dics,
                                              target_path,
                                              starti, endi))
            jobs.append(p)
        for i in range(len(jobs)):
            jobs[i].start()
        check_par_jobs(jobs)

    else:
        process_core_iterate(sta_ev_arr, input_dics, target_path,
                             0, len(sta_ev_arr))

# ###################### process_core_iterate #################################


def process_core_iterate(sta_ev_arr, input_dics, target_path, starti, endi):
    """
    running the process_unit for starti and endi defined by
    either serial or parallel mode
    :param sta_ev_arr:
    :param input_dics:
    :param target_path:
    :param starti:
    :param endi:
    :return:
    """
    import importlib
    try:
        process_unit = importlib.import_module(
            'obspyDMT.%s' % input_dics['pre_process'])
    except:
        from obspyDMT import __path__ as dmt_path
        sys.exit("\n\n%s.py DOES NOT EXIST at %s!"
                 % (input_dics['pre_process'], dmt_path))
        #process_unit = importlib.import_module(input_dics['pre_process'])
    for i in range(starti, endi):
        staev_ar = sta_ev_arr[i]
        station_id = '%s.%s.%s.%s' % (staev_ar[0], staev_ar[1],
                                      staev_ar[2], staev_ar[3])
        tr_add = os.path.join(target_path, 'raw', station_id)
        data_source = staev_ar[8]
        if input_dics['pre_process']:
            print('[%s/%s] start processing: %s'
                  % (i+1, len(sta_ev_arr), station_id))
            process_unit.process_unit(tr_add, target_path, input_dics, staev_ar)

# ###################### plot_unit ############################################


def plot_unit(input_dics, events):
    """
    this function re-direct the flow to the relevant plotting function
    :param input_dics:
    :param events:
    :return:
    """
    events = event_filter(events, input_dics)
    if input_dics['create_event_vtk']:
        vtk_generator(events)
    if input_dics['create_kml']:
        create_ev_sta_kml(input_dics, events)
    if input_dics['plot_seismicity']:
        plot_seismicity(input_dics, events)
    if input_dics['plot_ev'] or input_dics['plot_sta'] or \
            input_dics['plot_availability'] or input_dics['plot_ray']:
        plot_sta_ev_ray(input_dics, events)
    if input_dics['plot_waveform']:
        plot_waveform(input_dics, events)

# ##################### event_filter ##########################################


def event_filter(events, input_dics):
    """
    filtering events based on the inputs
    :param events:
    :param input_dics:
    :return:
    """
    del_index = []
    for ev in range(len(events)):
        if input_dics['dir_select']:
            if events[ev]['event_id'] not in input_dics['dir_select']:
                del_index.append(ev)
                continue
        if not plot_filter_event(input_dics, events[ev]):
            del_index.append(ev)
    del_index.sort(reverse=True)
    for di in del_index:
        del events[di]
    return events

# ##################### plot_filter_event #####################################


def plot_filter_event(input_dics, event_dic):
    """
    check whether the event can pass the criteria
    :param input_dics:
    :param event_dic:
    :return:
    """

    try:
        if not event_dic['datetime'] <= UTCDateTime(input_dics['max_date']):
            return False
        if not event_dic['datetime'] >= UTCDateTime(input_dics['min_date']):
            return False
        if not event_dic['magnitude'] < 0:
            if not event_dic['magnitude'] <= float(input_dics['max_mag']):
                return False
            if not event_dic['magnitude'] >= float(input_dics['min_mag']):
                return False
            if not event_dic['depth'] <= float(input_dics['max_depth']):
                return False
            if not event_dic['depth'] >= float(input_dics['min_depth']):
                return False
            if isinstance(input_dics['evlatmin'], float):

                if float(event_dic['longitude']) < 0:
                    event_dic_360 = 360. + float(event_dic['longitude'])
                else:
                    event_dic_360 = float(event_dic['longitude'])

                if float(input_dics['evlonmin']) < 0:
                    evlonmin_360 = 360. + float(input_dics['evlonmin'])
                else:
                    evlonmin_360 = float(input_dics['evlonmin'])

                if float(input_dics['evlonmax']) < 0:
                    evlonmax_360 = 360. + float(input_dics['evlonmax'])
                else:
                    evlonmax_360 = float(input_dics['evlonmax'])

                if not event_dic['latitude'] <= float(input_dics['evlatmax']):
                    return False
                if not event_dic_360 <= evlonmax_360:
                    return False
                if not event_dic['latitude'] >= float(input_dics['evlatmin']):
                    return False
                if not event_dic_360 >= evlonmin_360:
                    return False

        return True
    except Exception as error:
        return False

# ##################### plot_waveform #########################################


def plot_waveform(input_dics, events):
    """
    plot waveforms arranged by the epicentral distance
    :param input_dics:
    :param events:
    :return:
    """

    for ei in range(len(events)):
        target_path = locate(input_dics['datapath'], events[ei]['event_id'], num_matches=1)
        if len(target_path) == 0:
            continue
        if len(target_path) > 1:
            print("[LOCAL] more than one path was found for the event:")
            print(target_path)
            print("[INFO] use the first one:")
            target_path = target_path[0]
            print(target_path)
        else:
            print("[LOCAL] Path:")
            target_path = target_path[0]
            print(target_path)

        update_sta_ev_file(target_path, events[ei])
        sta_ev_arr = np.loadtxt(
            os.path.join(target_path, 'info', 'station_event'),
            delimiter=',', dtype=bytes, ndmin=2).astype(str)
        sta_ev_arr = sta_ev_arr.astype(object)
        del_index = []
        for sti in range(len(sta_ev_arr)):
            if not plot_filter_station(input_dics, sta_ev_arr[sti]):
                del_index.append(sti)
        del_index.sort(reverse=True)
        for di in del_index:
            sta_ev_arr[di] = np.delete(sta_ev_arr, (di), axis=0)

        for si in range(len(sta_ev_arr)):
            sta_id = sta_ev_arr[si, 0] + '.' + sta_ev_arr[si, 1] + '.' + \
                     sta_ev_arr[si, 2] + '.' + sta_ev_arr[si, 3]
            try:
                tr = read(os.path.join(target_path,
                                       input_dics['plot_dir_name'],
                                       sta_id))[0]
                time_diff = tr.stats.starttime - events[ei]['datetime']
                taxis = tr.times() + time_diff

                dist, azi, bazi = gps2DistAzimuth(events[ei]['latitude'],
                                                  events[ei]['longitude'],
                                                  float(sta_ev_arr[si, 4]),
                                                  float(sta_ev_arr[si, 5]))
                epi_dist = dist/111.194/1000.
                if input_dics['min_azi'] or input_dics['max_azi'] or \
                        input_dics['min_epi'] or input_dics['max_epi']:
                    if input_dics['min_epi']:
                        if epi_dist < input_dics['min_epi']:
                            continue
                    if input_dics['max_epi']:
                        if epi_dist > input_dics['max_epi']:
                            continue
                    if input_dics['min_azi']:
                        if azi < input_dics['min_azi']:
                            continue
                    if input_dics['max_azi']:
                        if azi > input_dics['max_azi']:
                            continue
                plt.plot(taxis, tr.normalize().data/2 + epi_dist, lw=0.9, c='k', alpha=0.3)
            except:
                continue

    plt.xlabel('Time (sec)', size=24)
    plt.ylabel('Distance (deg)', size=24)
    plt.xticks(size=18)
    plt.yticks(size=18)
    plt.tight_layout()
    if not input_dics['plot_save']:
        plt.savefig(os.path.join(os.path.curdir, 'waveforms.png'), dpi=300)
    else:
        plt.savefig(os.path.join(input_dics['plot_save']), dpi=300)
    if not input_dics['show_no_plot']:
        plt.show()

# ##################### plot_sta_ev_ray ###############################


def plot_sta_ev_ray(input_dics, events):
    """
    plot stations, events and ray paths on a map using basemap.
    :param input_dics:
    :param events:
    :return:
    """

    import cartopy.crs as ccrs
    import cartopy.feature as cfeature


    fig = plt.figure(figsize=(20, 10))
    plt_stations = input_dics['plot_sta']
    plt_availability = input_dics['plot_availability']
    plt_events = input_dics['plot_ev']
    plt_ray_path = input_dics['plot_ray']

    if input_dics['evlatmin'] is None:
        evlatmin = -90
        evlatmax = +90
        evlonmin = -180
        evlonmax = +180
        glob_map = True
    else:
        evlatmin = input_dics['evlatmin']
        evlatmax = input_dics['evlatmax']
        evlonmin = input_dics['evlonmin']
        evlonmax = input_dics['evlonmax']
        glob_map = False

    if not glob_map:
        map_ax = fig.add_subplot(111, projection=ccrs.PlateCarree())
        map_ax.gridlines()
        map_ax.set_extent([evlonmin, evlonmax, evlatmin, evlatmax], crs=ccrs.Geodetic())
        width_beach = 1
        width_station = 30
    elif glob_map:
        map_ax = fig.add_subplot(111, projection=ccrs.Robinson(central_longitude=input_dics["plot_lon0"]))
        map_ax.set_global()
        map_ax.gridlines()
        width_beach = 1e6
        width_station = 30
    else:
        sys.exit('[ERROR] can not continue, error:\n%s' % input_dics)

    if input_dics['plot_style'] == 'bluemarble':
        print("[WARNING] bluemarble is not supported in v2.2.6, "
              "using 'a downsampled version of the Natural Earth shaded relief raster'.")
        map_ax.stock_img()
    elif input_dics['plot_style'] == 'etopo':
        print("[WARNING] etopo is not supported in v2.2.6, "
              "using 'a downsampled version of the Natural Earth shaded relief raster'.")
        map_ax.stock_img()
    elif input_dics['plot_style'] == 'shadedrelief':
        print("[WARNING] 'a downsampled version of the Natural Earth shaded relief raster' will be used.")
        map_ax.stock_img()
    else:
        map_ax.add_feature(cfeature.LAND, facecolor='#bfbfbf')
        # map_ax.coastlines()

    for ei in range(len(events)):
        if plt_events:
            if input_dics["plot_focal"]:
                print("\n[WARNING] Plotting beachball(s) is under construction! Change to simple point plot.\n")
                input_dics["plot_focal"] = False
            if input_dics['plot_focal']:
                if not events[ei]['focal_mechanism']:
                    print('[ERROR] moment tensor does not exist!')
                    continue
                x, y = (events[ei]['longitude'], events[ei]['latitude'])
                focmecs = [float(events[ei]['focal_mechanism'][0]),
                           float(events[ei]['focal_mechanism'][1]),
                           float(events[ei]['focal_mechanism'][2]),
                           float(events[ei]['focal_mechanism'][3]),
                           float(events[ei]['focal_mechanism'][4]),
                           float(events[ei]['focal_mechanism'][5])]
                try:
                    ax = plt.gca()
                    b = Beach(focmecs, xy=(x, y), facecolor='blue',
                              width=width_beach, linewidth=1, alpha=0.85)
                    b.set_zorder(10)
                    ax.add_collection(b)
                except Exception as error:
                    print("[WARNING: %s -- %s]" % (error, focmecs))
                    continue
            else:
                x, y = (events[ei]['longitude'], events[ei]['latitude'])
                magnitude = float(events[ei]['magnitude'])
                plt.scatter(x, y, color="blue", s=10*magnitude,
                            edgecolors='none', marker="o",
                            zorder=5, alpha=0.65, 
                            transform=ccrs.Geodetic())
        if plt_stations or plt_availability or plt_ray_path:
            target_path = locate(input_dics['datapath'],
                                 events[ei]['event_id'], num_matches=1)
            if len(target_path) == 0:
                continue
            if len(target_path) > 1:
                print("[LOCAL] more than one path was found for the event:")
                print(target_path)
                print("[INFO] use the first one:")
                target_path = target_path[0]
                print(target_path)
            else:
                print("[LOCAL] Path:")
                target_path = target_path[0]
                print(target_path)

            update_sta_ev_file(target_path, events[ei])
            if not input_dics['plot_availability']:
                sta_ev_arr = np.loadtxt(os.path.join(target_path,
                                                     'info', 'station_event'),
                                        delimiter=',', dtype=bytes, ndmin=2).astype(str)
            else:
                sta_ev_arr = np.loadtxt(os.path.join(target_path,
                                                     'info',
                                                     'availability.txt'),
                                        delimiter=',', dtype=bytes, ndmin=2).astype(str)
            sta_ev_arr = sta_ev_arr.astype(object)

            if events[ei]['magnitude'] > 0:
                del_index = []
                for sti in range(len(sta_ev_arr)):

                    if not plot_filter_station(input_dics, sta_ev_arr[sti]):
                        del_index.append(sti)

                    dist, azi, bazi = gps2DistAzimuth(
                        events[ei]['latitude'],
                        events[ei]['longitude'],
                        float(sta_ev_arr[sti, 4]),
                        float(sta_ev_arr[sti, 5]))

                    epi_dist = dist/111.194/1000.
                    if input_dics['min_azi'] or input_dics['max_azi'] or \
                            input_dics['min_epi'] or input_dics['max_epi']:
                        if input_dics['min_epi']:
                            if epi_dist < input_dics['min_epi']:
                                del_index.append(sti)
                        if input_dics['max_epi']:
                            if epi_dist > input_dics['max_epi']:
                                del_index.append(sti)
                        if input_dics['min_azi']:
                            if azi < input_dics['min_azi']:
                                del_index.append(sti)
                        if input_dics['max_azi']:
                            if azi > input_dics['max_azi']:
                                del_index.append(sti)

                del_index = list(set(del_index))
                del_index.sort(reverse=True)
                for di in del_index:
                    sta_ev_arr = np.delete(sta_ev_arr, (di), axis=0)

        if plt_stations or plt_availability:
            if len(sta_ev_arr) > 0:
                x, y = (sta_ev_arr[:, 5], sta_ev_arr[:, 4])
                plt.scatter(x.astype(np.float), y.astype(np.float),
                            color='red', s=width_station,
                            edgecolors='none', marker='v',
                            zorder=4, alpha=0.9,
                            transform=ccrs.Geodetic())

        if plt_ray_path:
            for si in range(len(sta_ev_arr)):
                plt.plot([float(events[ei]['longitude']), float(sta_ev_arr[si][5])],
                         [float(events[ei]['latitude']), float(sta_ev_arr[si][4])], 
                         color='k', alpha=0.1, 
                         transform=ccrs.Geodetic())
                         
    if not input_dics['plot_save']:
        plt.savefig(os.path.join(os.path.curdir, 'event_station.png'))
    else:
        plt.savefig(os.path.join(input_dics['plot_save']))
    if not input_dics['show_no_plot']:
        plt.show()

# ##################### plot_seismicity #######################################


def plot_seismicity(input_dics, events):
    """
    create a seismicity map with some basic statistical analysis on the results
    :param input_dics:
    :param events:
    :return:
    """
    print('\n==============')
    print('Seismicity map')
    print('==============\n')

    from mpl_toolkits.basemap import Basemap

    # plt.rc('font', family='serif')
    if not len(events) > 0:
        print("[WARNING] no event passed the given criteria!")
        print("[WARNING] can not create any seismicity map.")
        return

    if input_dics['evlatmin'] is None:
        input_dics['evlatmin'] = -90
        input_dics['evlatmax'] = +90
        input_dics['evlonmin'] = -180
        input_dics['evlonmax'] = +180
        map_proj = 'cyl'
    else:
        map_proj = 'cyl'

    # set-up the map
    m = Basemap(projection=map_proj,
                llcrnrlat=input_dics['evlatmin'],
                urcrnrlat=input_dics['evlatmax'],
                llcrnrlon=input_dics['evlonmin'],
                urcrnrlon=input_dics['evlonmax'],
                lon_0=input_dics['plot_lon0'],
                resolution='l')

    parallels = np.arange(-90, 90, 30.)
    m.drawparallels(parallels, fontsize=24)
    meridians = np.arange(-180., 180., 60.)
    m.drawmeridians(meridians, fontsize=24)

    if input_dics['plot_style'] == 'bluemarble':
        m.bluemarble(scale=0.5)
    elif input_dics['plot_style'] == 'etopo':
        m.etopo(scale=0.5)
    elif input_dics['plot_style'] == 'shadedrelief':
        m.shadedrelief(scale=0.1)
    else:
        m.fillcontinents()

    # defining labels
    x_ev, y_ev = m(-360, 0)
    m.scatter(x_ev, y_ev, 20, color='red', marker="o",
              edgecolor="black", zorder=10, label='0-70km')
    m.scatter(x_ev, y_ev, 20, color='green', marker="o",
              edgecolor="black", zorder=10, label='70-300km')
    m.scatter(x_ev, y_ev, 20, color='blue', marker="o",
              edgecolor="black", zorder=10, label='300< km')

    m.scatter(x_ev, y_ev, 10, color='white', marker="o",
              edgecolor="black", zorder=10, label='< 4.0')
    m.scatter(x_ev, y_ev, 40, color='white', marker="o",
              edgecolor="black", zorder=10, label='4.0-5.0')
    m.scatter(x_ev, y_ev, 70, color='white', marker="o",
              edgecolor="black", zorder=10, label='5.0-6.0')
    m.scatter(x_ev, y_ev, 100, color='white', marker="o",
              edgecolor="black", zorder=10, label='6.0<=')

    ev_dp_all = []
    ev_mag_all = []
    ev_info_ar = np.array([])
    plot_focal_mechanism = False
    for ev in events:
        x_ev, y_ev = m(float(ev['longitude']), float(ev['latitude']))
        ev_dp_all.append(abs(float(ev['depth'])))
        ev_mag_all.append(abs(float(ev['magnitude'])))

        if abs(float(ev['depth'])) < 70.0:
            color = 'red'
        elif 70.0 <= abs(float(ev['depth'])) < 300.0:
            color = 'green'
        elif 300.0 <= abs(float(ev['depth'])):
            color = 'blue'

        if float(ev['magnitude']) < 4.0:
            size = 10
        elif 4.0 <= float(ev['magnitude']) < 5.0:
            size = 40
        elif 5.0 <= float(ev['magnitude']) < 6.0:
            size = 70
        elif 6.0 <= float(ev['magnitude']):
            size = 100

        if ev['focal_mechanism']:
            plot_focal_mechanism = True
            f1 = ev['focal_mechanism'][0]
            f2 = ev['focal_mechanism'][1]
            f3 = ev['focal_mechanism'][2]
            f4 = ev['focal_mechanism'][3]
            f5 = ev['focal_mechanism'][4]
            f6 = ev['focal_mechanism'][5]
        else:
            f1 = False
            f2 = False
            f3 = False
            f4 = False
            f5 = False
            f6 = False
        if np.size(ev_info_ar) < 1:
            ev_info_ar = np.append(ev_info_ar,
                                   [float(ev['depth']), float(x_ev),
                                    float(y_ev), size, color,
                                    f1, f2, f3, f4, f5, f6])
        else:
            ev_info_ar = np.vstack((ev_info_ar,
                                    [float(ev['depth']), float(x_ev),
                                     float(y_ev), size, color,
                                     f1, f2, f3, f4, f5, f6]))

    if np.shape(ev_info_ar)[0] == np.size(ev_info_ar):
        ev_info_ar = np.reshape(ev_info_ar, (1, 11))
    else:
        ev_info_ar = sorted(ev_info_ar,
                            key=lambda ev_info_iter: float(ev_info_iter[0]))

    for ev in ev_info_ar:
        m.scatter(float(ev[1]), float(ev[2]), float(ev[3]),
                  color=ev[4], marker="o", edgecolor='k', zorder=10)

    plt.legend(bbox_to_anchor=(1.01, 1), loc=2, borderaxespad=0.)

    if plot_focal_mechanism:
        plt.figure()

        m = Basemap(projection=map_proj,
                    llcrnrlat=input_dics['evlatmin'],
                    urcrnrlat=input_dics['evlatmax'],
                    llcrnrlon=input_dics['evlonmin'],
                    urcrnrlon=input_dics['evlonmax'],
                    lon_0=input_dics['plot_lon0'],
                    resolution='l')

        parallels = np.arange(-90, 90, 30.)
        m.drawparallels(parallels, fontsize=24)
        meridians = np.arange(-180., 180., 60.)
        m.drawmeridians(meridians, fontsize=24)

        if input_dics['plot_style'] == 'bluemarble':
            m.bluemarble(scale=0.5)
        elif input_dics['plot_style'] == 'etopo':
            m.etopo(scale=0.5)
        elif input_dics['plot_style'] == 'shadedrelief':
            m.shadedrelief(scale=0.1)
        else:
            m.fillcontinents()

        for evfoc in ev_info_ar:
            focmec = False
            try:
                ax = plt.gca()
                focmec = (float(evfoc[5]), float(evfoc[6]), float(evfoc[7]),
                          float(evfoc[8]), float(evfoc[9]), float(evfoc[10]))
                b = Beach(focmec, xy=(float(evfoc[1]), float(evfoc[2])),
                          facecolor=evfoc[4], width=float(evfoc[3])/100.,
                          linewidth=1, alpha=0.85)
                b.set_zorder(10)
                ax.add_collection(b)
            except Exception as error:
                print('[EXCEPTION] focal mechanism:')
                print(focmec)
                print('[EXCEPTION] error: %s' % error)

    if len(events) > 0:
        plt.figure()
        hn, hbins, hpatches = \
            plt.hist(ev_dp_all, input_dics['depth_bins_seismicity'],
                     facecolor='g', edgecolor='k', lw=3, alpha=0.5, log=True)

        plt.xlabel('Depth', size=32, weight='bold')
        plt.ylabel('#Events (log)', size=32, weight='bold')
        plt.yscale('log')
        plt.ylim(ymin=0.2)
        plt.xticks(size=24, weight='bold')
        plt.yticks(size=24, weight='bold')
        plt.tight_layout()
        plt.grid(True)

        plt.figure()
        hn, hbins, hpatches = \
            plt.hist(ev_mag_all,
                     bins=np.linspace(int(float(input_dics['min_mag'])),
                                      int(float(input_dics['max_mag'])),
                                      (int(float(input_dics['max_mag'])) -
                                       int(float(input_dics['min_mag'])))*2+1),
                     facecolor='g', edgecolor='k', lw=3, alpha=0.5, log=True)

        plt.xlabel('Magnitude', size=32, weight='bold')
        plt.ylabel('#Events (log)', size=32, weight='bold')
        plt.yscale('log')
        plt.ylim(ymin=0.2)
        plt.xticks(size=24, weight='bold')
        plt.yticks(size=24, weight='bold')
        plt.tight_layout()
        plt.grid(True)

    if not input_dics['plot_save']:
        plt.savefig(os.path.join(os.path.curdir, 'seismicity.png'))
    else:
        plt.savefig(os.path.join(input_dics['plot_save']))
    if not input_dics['show_no_plot']:
        plt.show()

# ##################### vtk_generator ###################################


def vtk_generator(events, vtk_output='events'):
    """
    VTK generator for events
    :param events:
    :param vtk_output:
    :return:
    """
    counter = 0
    xyz = []
    for i in range(len(events)):
        elat = events[i]['latitude']*np.pi/180.
        elon = events[i]['longitude']*np.pi/180.
        edp = events[i]['depth']
        x = (6371-edp) * np.cos(elat) * np.cos(elon)
        y = (6371-edp) * np.cos(elat) * np.sin(elon)
        z = (6371-edp) * np.sin(elat)
        print
        xyz.append("%s %s %s" % (x, y, z))
        counter += 1
    fio = open(vtk_output + '.vtk', 'wt')
    fio.writelines('# vtk DataFile Version 3.0\n')
    fio.writelines('vtk output\n')
    fio.writelines('ASCII\n')
    fio.writelines('DATASET UNSTRUCTURED_GRID\n')
    fio.writelines('POINTS %i float\n' % counter)
    for i in range(len(xyz)):
        fio.writelines('%s\n' % xyz[i])

    fio.writelines('\n')
    fio.writelines('CELLS %i %i\n' % (counter, counter*2))
    for i in range(len(xyz)):
        fio.writelines('1 %s\n' % i)

    fio.writelines('\n')
    fio.writelines('CELL_TYPES %i\n' % counter)
    for i in range(len(xyz)):
        fio.writelines('1\n')
    fio.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -------------------------------------------------------------------
#   Filename:  kml_handler.py
#   Purpose:   handling KML format in obspyDMT
#   Author:    Kasra Hosseini
#   Email:     kasra.hosseinizad@earth.ox.ac.uk
#   License:   GNU Lesser General Public License, Version 3
# -------------------------------------------------------------------

# -----------------------------------------------------------------------
# ----------------Import required Modules (Python and Obspy)-------------
# -----------------------------------------------------------------------
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import matplotlib.pyplot as plt
import numpy as np
try:
    from obspy.imaging.beachball import beachball as Beachball
except:
    from obspy.imaging.beachball import Beachball
try:
    from obspy.geodetics import locations2degrees
except:
    from obspy.core.util import locations2degrees
try:
    from obspy.geodetics.base import gps2dist_azimuth as gps2DistAzimuth
except:
    try:
        from obspy.geodetics import gps2DistAzimuth
    except:
        from obspy.core.util import gps2DistAzimuth
import os
import sys

from .data_handler import update_sta_ev_file
from .utility_codes import locate, plot_filter_station

# ##################### create_ev_sta_kml ###############################


def create_ev_sta_kml(input_dics, events):
    """
    create event/station/ray in KML format readable by Google-Earth
    :param input_dics:
    :param events:
    :return:
    """
    try:
        from pykml.factory import KML_ElementMaker as KML
        from lxml import etree
    except:
        sys.exit('[ERROR] pykml should be installed first!')

    if not os.path.isdir('kml_dir'):
        os.mkdir('kml_dir')
    else:
        print('[INFO] kml_dir already exists!')

    # create a document element with multiple label style
    kmlobj = KML.kml(KML.Document())
    counter = 0
    for ei in range(len(events)):
        print(events[ei]['event_id'])
        counter += 1

        ev_date = '%04i/%02i/%02i-%02i:%02i:%02i' \
                  % (events[ei]['datetime'].year,
                     events[ei]['datetime'].month,
                     events[ei]['datetime'].day,
                     events[ei]['datetime'].hour,
                     events[ei]['datetime'].minute,
                     events[ei]['datetime'].second
                     )

        if input_dics['plot_focal']:
            try:
                focmecs = [float(events[ei]['focal_mechanism'][0]),
                           float(events[ei]['focal_mechanism'][1]),
                           float(events[ei]['focal_mechanism'][2]),
                           float(events[ei]['focal_mechanism'][3]),
                           float(events[ei]['focal_mechanism'][4]),
                           float(events[ei]['focal_mechanism'][5])]
            except:
                print("WARNING: 'focal_mechanism' does not exist!")
                focmecs = [1, 1, 1, 0, 0, 0]
        else:
            focmecs = [1, 1, 1, 0, 0, 0]

        if 0 <= events[ei]['depth'] < 70:
            event_color = 'red'
        elif 70 <= events[ei]['depth'] < 300:
            event_color = 'green'
        else:
            event_color = 'blue'

        try:
            Beachball(focmecs,
                      outfile=os.path.join(
                          'kml_dir', events[ei]['event_id'] + '.png'),
                      facecolor=event_color,
                      edgecolor=event_color)
        except Exception as error:
            print(error)
            print(focmecs)
            continue
        plt.close()
        plt.clf()

        if input_dics['plot_ev'] or input_dics['plot_ray']:
            kmlobj.Document.append(
                KML.Style(
                    KML.IconStyle(
                        KML.Icon(KML.href(os.path.join(
                            events[ei]['event_id'] + '.png')),
                        ),
                        KML.scale(events[ei]['magnitude']/2.),
                        KML.heading(0.0),
                    ),
                    id='beach_ball_%i' % counter
                ),
            )

            kmlobj.Document.append(
                KML.Placemark(
                    KML.name(events[ei]['event_id']),
                    KML.ExtendedData(
                        KML.Data(
                            KML.value('%s' % events[ei]['event_id']),
                            name='event_id'
                        ),
                        KML.Data(
                            KML.value('%s' % events[ei]['magnitude']),
                            name='magnitude'
                        ),
                        KML.Data(
                            KML.value('%s' % ev_date),
                            name='datetime'
                        ),
                        KML.Data(
                            KML.value('%s' % events[ei]['depth']),
                            name='depth'
                        ),
                        KML.Data(
                            KML.value('%s' % events[ei]['latitude']),
                            name='latitude'
                        ),
                        KML.Data(
                            KML.value('%s' % events[ei]['longitude']),
                            name='longitude'
                        ),
                    ),
                    KML.styleUrl('#beach_ball_%i' % counter),
                    # KML.Point(KML.coordinates(events[ei]['longitude'], ',',
                    #                           events[ei]['latitude'], ',',
                    #                           700000 -
                    #                           abs(events[ei]['depth']*1000)),
                    #           KML.altitudeMode('absolute')
                    #           ),
                    KML.Point(KML.coordinates(events[ei]['longitude'], ',',
                                              events[ei]['latitude'], ',',
                                              0.),
                              KML.altitudeMode('absolute')
                              ),
                ),
            )

        if input_dics['plot_sta'] or input_dics['plot_ray']:
            target_path = locate(input_dics['datapath'],
                                 events[ei]['event_id'], num_matches=1)
            if len(target_path) < 1:
                continue
            if len(target_path) > 1:
                print("[LOCAL] more than one path was found for the event:")
                print(target_path)
                print("[INFO] use the first one:")
                target_path = target_path[0]
                print(target_path)
            else:
                print("[LOCAL] Path:")
                target_path = target_path[0]
                print(target_path)

            update_sta_ev_file(target_path, events[ei])
            sta_ev_arr = np.loadtxt(os.path.join(target_path,
                                                 'info', 'station_event'),
                                    delimiter=',', dtype=bytes, ndmin=2).astype(np.str)
            sta_ev_arr = sta_ev_arr.astype(np.object)
            del_index = []
            for sti in range(len(sta_ev_arr)):

                if not plot_filter_station(input_dics, sta_ev_arr[sti]):
                    del_index.append(sti)

                dist, azi, bazi = gps2DistAzimuth(events[ei]['latitude'],
                                                  events[ei]['longitude'],
                                                  float(sta_ev_arr[sti, 4]),
                                                  float(sta_ev_arr[sti, 5]))

                epi_dist = dist/111.194/1000.
                if input_dics['min_azi'] or input_dics['max_azi'] or \
                        input_dics['min_epi'] or input_dics['max_epi']:
                    if input_dics['min_epi']:
                        if epi_dist < input_dics['min_epi']:
                            del_index.append(sti)
                    if input_dics['max_epi']:
                        if epi_dist > input_dics['max_epi']:
                            del_index.append(sti)
                    if input_dics['min_azi']:
                        if azi < input_dics['min_azi']:
                            del_index.append(sti)
                    if input_dics['max_azi']:
                        if azi > input_dics['max_azi']:
                            del_index.append(sti)

            del_index = list(set(del_index))
            del_index.sort(reverse=True)
            for di in del_index:
                sta_ev_arr = np.delete(sta_ev_arr, (di), axis=0)

            kmlobj.Document.append(
                KML.Style(
                    KML.IconStyle(
                        KML.scale(2.5),
                        KML.heading(0.0),
                    ),
                    id='station'
                ),
            )
            kmlobj.Document.append(
                KML.Style(
                    KML.LineStyle(
                        KML.width(1.0),
                        # KML.color('ff33ccff'),
                        KML.color('2333ccff'),
                    ),
                    id='great_circle_distance'
                ),
            )
            for sti in sta_ev_arr:
                dist, azi, bazi = gps2DistAzimuth(events[ei]['latitude'],
                                                  events[ei]['longitude'],
                                                  float(sti[4]),
                                                  float(sti[5]))
                epi_dist = dist/111.194/1000.
                sta_id = '%s.%s.%s.%s' % (sti[0], sti[1], sti[2], sti[3])
                kmlobj.Document.append(
                    KML.Placemark(
                        KML.name(sta_id),
                        KML.ExtendedData(
                            KML.Data(
                                KML.value('%s' % sta_id),
                                name='StationID'
                            ),
                            KML.Data(
                                KML.value('%s' % epi_dist),
                                name='Distance'
                            ),
                            KML.Data(
                                KML.value('%s' % sti[6]),
                                name='Elevation'
                            ),
                            KML.Data(
                                KML.value('%s' % sti[7]),
                                name='Depth'
                            ),
                            KML.Data(
                                KML.value('%s' % bazi),
                                name='Back-Azimuth'
                            ),
                            KML.Data(
                                KML.value('%s' % azi),
                                name='Azimuth'
                            ),
                            KML.Data(
                                KML.value('%s' % events[ei]['event_id']),
                                name='EventID'
                            ),
                            KML.Data(
                                KML.value('%s' % sti[8]),
                                name='Source'
                            ),
                        ),
                        KML.styleUrl('station'),
                        KML.Point(KML.coordinates(
                            float(sti[5]), ',', float(sti[4]))
                        ),
                    ),
                )
                if input_dics['plot_ray']:
                    kmlobj.Document.append(
                        KML.Placemark(
                            KML.name(sta_id),
                            KML.ExtendedData(
                                KML.Data(
                                    KML.value('%s' % sta_id),
                                    name='StationID'
                                ),
                            ),
                            KML.styleUrl('great_circle_distance'),
                            KML.LineString(KML.coordinates(
                                '%s,%s,0\n'
                                '%s,%s,0' % (float(sti[5]),
                                             float(sti[4]),
                                             events[ei]['longitude'],
                                             events[ei]['latitude'])),
                                KML.tessellate(1)),
                        ),
                    )
    kml_outfile = open(os.path.join('kml_dir', 'kml_output.kml'), 'wb')
    kml_outfile.write(etree.tostring(kmlobj, pretty_print=True))
    sys.exit('[INFO] KML file is stored in ./kml_dir!')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -------------------------------------------------------------------
#   Filename:  data_handler.py
#   Purpose:   handling data (waveform/response) in obspyDMT
#   Author:    Kasra Hosseini
#   Email:     kasra.hosseinizad@earth.ox.ac.uk
#   License:   GNU Lesser General Public License, Version 3
# -------------------------------------------------------------------

# -----------------------------------------------------------------------
# ----------------Import required Modules (Python and Obspy)-------------
# -----------------------------------------------------------------------
from __future__ import print_function
from datetime import datetime
import fileinput
import glob
import multiprocessing
import numpy as np
try:
    from obspy.clients.fdsn import Client as Client_fdsn
except:
    from obspy.fdsn import Client as Client_fdsn
try:
    from obspy.clients.fdsn import RoutingClient
except:
    print("[WARNING] RoutingClient could not be imported.")
from obspy.clients.syngine import Client as Client_syngine
try:
    from obspy.geodetics.base import gps2dist_azimuth as gps2DistAzimuth
except:
    try:
        from obspy.geodetics import gps2DistAzimuth
    except:
        from obspy.core.util import gps2DistAzimuth
from obspy import read_inventory
import os
import pickle

from .utility_codes import calculate_time_phase, getFolderSize
from .utility_codes import geocen_calc

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

# ##################### get_data ###############################


def get_data(stas_avail, event, input_dics, info_event):
    """
    get the waveform/response from FDSN
    :param stas_avail:
    :param event:
    :param input_dics:
    :param info_event:
    :return:
    """
    t_wave_1 = datetime.now()

    req_clients = np.unique(stas_avail[:, 8])
    print("[INFO] requested data sources:")
    for req_c in req_clients:
        print(req_c, end=' ')
    print('\n')

    if input_dics['test']:
        stas_avail = stas_avail[0:input_dics['test_num']]

    if input_dics['req_parallel']:
        par_jobs = []
        for req_cli in req_clients:
            st_avail = stas_avail[stas_avail[:, 8] == req_cli]

            p = multiprocessing.Process(target=fdsn_waveform,
                                        args=(st_avail, event,
                                                input_dics, req_cli,
                                                info_event))
            par_jobs.append(p)

        sub_par_jobs = []
        for l in range(len(par_jobs)):
            counter = len(req_clients)
            while counter >= len(req_clients):
                counter = 0
                for ll in range(len(sub_par_jobs)):
                    if par_jobs[sub_par_jobs[ll]].is_alive():
                        counter += 1
            par_jobs[l].start()
            sub_par_jobs.append(l)

        counter = len(req_clients)
        while counter > 0:
            counter = 0
            for ll in range(len(par_jobs)):
                if par_jobs[ll].is_alive():
                    counter += 1
    else:
        for req_cli in req_clients:
            st_avail = stas_avail[stas_avail[:, 8] == req_cli]
            fdsn_waveform(st_avail, event, input_dics, req_cli, info_event)

    print("\n========================")
    print("DONE with Event: %s" % event['event_id'])
    print("Time: %s" % (datetime.now() - t_wave_1))
    print("========================")

# ##################### fdsn_waveform ###############################


def fdsn_waveform(stas_avail, event, input_dics, req_cli, info_event):
    """
    get Waveforms, StationXML files and meta-data from FDSN
    :param stas_avail:
    :param event:
    :param input_dics:
    :param req_cli:
    :param info_event:
    :return:
    """
    eventpath = os.path.join(input_dics['datapath'])
    target_path = os.path.join(eventpath, event['event_id'])

    if input_dics['bulk']:
        try:
            fdsn_bulk_request(target_path, req_cli, input_dics)
        except Exception as error:
            print('[WARNING] %s' % error)
        print('DONE')

        # Following parameter is set to 'N' to avoid
        # retrieving the waveforms twice
        # When using bulk requests, waveforms are retreived in bulk
        # but not response/StationXML files and not metadata
        input_dics['waveform'] = False
        print('%s bulkdataselect request is done for event: %s' \
              % (req_cli, target_path))

    fdsn_serial_parallel(stas_avail, event, input_dics, target_path,
                         req_cli, info_event)

# ##################### fdsn_serial_parallel ##################################


def fdsn_serial_parallel(stas_avail, event, input_dics, target_path,
                         req_cli, info_event):
    """
    retrieving data from FDSN
    :param stas_avail:
    :param event:
    :param input_dics:
    :param target_path:
    :param req_cli:
    :param info_event:
    :return:
    """
    print('%s -- event: %s' % (req_cli, target_path))

    if req_cli.lower() in ["iris-federator", "eida-routing"]:
        client_fdsn = RoutingClient(req_cli)
    else:
        client_fdsn = Client_fdsn(base_url=req_cli,
                                  user=input_dics['username_fdsn'],
                                  password=input_dics['password_fdsn'])
                                  #debug=True)
    client_syngine = Client_syngine()

    if input_dics['req_parallel']:
        if input_dics['password_fdsn']:
            print("[INFO] Restricted data from %s" % req_cli)
            print("[WARNING] parallel retrieving is now possible!")
            print("[WARNING] serial retrieving is activated!")
            # num_req_np = 1
            num_req_np = input_dics['req_np']
        else:
            num_req_np = input_dics['req_np']
        par_jobs = []
        st_counter = 0
        for st_avail in stas_avail:
            st_counter += 1
            info_station = '[%s-%s/%s]' % (info_event, st_counter,
                                           len(stas_avail))
            p = multiprocessing.Process(target=fdsn_download_core,
                                        args=(st_avail, event,
                                              input_dics, target_path,
                                              client_fdsn, client_syngine,
                                              req_cli,
                                              info_station))
            par_jobs.append(p)

        sub_par_jobs = []
        for l in range(len(par_jobs)):
            counter = num_req_np
            while counter >= num_req_np:
                counter = 0
                for ll in range(len(sub_par_jobs)):
                    if par_jobs[sub_par_jobs[ll]].is_alive():
                        counter += 1
            par_jobs[l].start()
            sub_par_jobs.append(l)

        counter = num_req_np
        while counter > 0:
            counter = 0
            for ll in range(len(par_jobs)):
                if par_jobs[ll].is_alive():
                    counter += 1
    else:
        st_counter = 0
        for st_avail in stas_avail:
            st_counter += 1
            info_station = '[%s-%s/%s]' % (info_event, st_counter,
                                           len(stas_avail))
            fdsn_download_core(st_avail, event, input_dics, target_path,
                               client_fdsn, client_syngine,
                               req_cli, info_station)

    update_sta_ev_file(target_path, event)

    if input_dics['bulk']:
        input_dics['waveform'] = True
        sta_saved_path = glob.glob(os.path.join(target_path, 'raw', '*.*.*.*'))
        print('\n[INFO] adjusting the station_event file for bulk request...', end='')

        sta_saved_list = []
        for sta_num in range(len(sta_saved_path)):
            sta_saved_list.append(os.path.basename(sta_saved_path[sta_num]))

        sta_ev_new = []
        for line in fileinput.FileInput(
                os.path.join(target_path, 'info', 'station_event')):
            line_split = line.split(',')
            if not '%s.%s.%s.%s' \
                    % (line_split[0], line_split[1], line_split[2],
                       line_split[3]) in sta_saved_list:
                pass
            else:
                sta_ev_new.append(line)

        file_staev_open = open(os.path.join(target_path, 'info',
                                            'station_event'), 'wt')
        file_staev_open.writelines(sta_ev_new)
        file_staev_open.close()
        print('DONE')

# ##################### fdsn_download_core ##################################


def fdsn_download_core(st_avail, event, input_dics, target_path,
                       client_fdsn, client_syngine, req_cli, info_station):
    """
    downloading the waveforms, reponse files (StationXML) and metadata
    this function should be normally called by some higher-level functions
    :param st_avail:
    :param event:
    :param input_dics:
    :param target_path:
    :param client_fdsn:
    :param client_syngine:
    :param req_cli:
    :param info_station:
    :return:
    """
    dummy = 'initializing'

    t11 = datetime.now()
    identifier = 0
    st_id = '%s.%s.%s.%s' % (st_avail[0], st_avail[1],
                             st_avail[2], st_avail[3])
    try:
        if st_avail[2] == '--' or st_avail[2] == '  ':
                st_avail[2] = ''

        if input_dics['cut_time_phase']:
            t_start, t_end = calculate_time_phase(event, st_avail)
        else:
            t_start = event['t1']
            t_end = event['t2']

        if input_dics['min_azi'] or input_dics['max_azi'] or \
                input_dics['min_epi'] or input_dics['max_epi']:
            dist, azi, bazi = gps2DistAzimuth(event['latitude'],
                                              event['longitude'],
                                              float(st_avail[4]),
                                              float(st_avail[5]))
            epi_dist = dist/111.194/1000.
            if input_dics['min_epi']:
                if epi_dist < input_dics['min_epi']:
                    raise Exception('%s out of epi range: %s'
                                    % (st_id, epi_dist))
            if input_dics['max_epi']:
                if epi_dist > input_dics['max_epi']:
                    raise Exception('%s out of epi range: %s'
                                    % (st_id, epi_dist))
            if input_dics['min_azi']:
                if azi < input_dics['min_azi']:
                    raise Exception('%s outo f Azimuth range: %s'
                                    % (st_id, azi))
            if input_dics['max_azi']:
                if azi > input_dics['max_azi']:
                    raise Exception('%s out of Azimuth range: %s'
                                    % (st_id, azi))

        if input_dics['waveform']:
            dummy = 'waveform'
            if (not os.path.isfile(os.path.join(target_path, 'raw', st_id)))\
                    or input_dics['force_waveform']:
                waveforms_path2write = os.path.join(target_path, 'raw', st_id)
                if req_cli.lower() in ["iris-federator", "eida-routing"]:
                    dl_waveform = \
                        client_fdsn.get_waveforms(
                            network=st_avail[0],
                            station=st_avail[1],
                            location=st_avail[2],
                            channel=st_avail[3],
                            starttime=t_start,
                            endtime=t_end)
                    if len(dl_waveform) > 0:
                        dl_waveform.write(waveforms_path2write, format='mseed')
                    else:
                        raise Exception

                else:
                    client_fdsn.get_waveforms(network=st_avail[0],
                                              station=st_avail[1],
                                              location=st_avail[2],
                                              channel=st_avail[3],
                                              starttime=t_start,
                                              endtime=t_end,
                                              filename=waveforms_path2write)
                identifier += 10
                print('%s -- %s -- saving waveform for: %s  ---> DONE' \
                      % (info_station, req_cli, st_id))
            else:
                identifier += 1

        if input_dics['response']:
            dummy = 'response'
            if (not os.path.isfile(os.path.join(target_path, 'resp',
                                                'STXML.' + st_id))) \
                    or input_dics['force_response']:
                resp_path2write = os.path.join(target_path, 'resp', 'STXML.%s' % st_id)
                if req_cli.lower() in ["iris-federator", "eida-routing"]:
                    dl_response = \
                        client_fdsn.get_stations(
                            network=st_avail[0],
                            station=st_avail[1],
                            location=st_avail[2],
                            channel=st_avail[3],
                            starttime=t_start,
                            endtime=t_end,
                            level='response')
                    if len(dl_response) > 0:
                        dl_response.write(resp_path2write, format="stationxml")
                    else:
                        raise Exception

                else:
                    client_fdsn.get_stations(network=st_avail[0],
                                             station=st_avail[1],
                                             location=st_avail[2],
                                             channel=st_avail[3],
                                             starttime=t_start,
                                             endtime=t_end,
                                             filename=resp_path2write,
                                             level='response')
                identifier += 100
                print("%s -- %s -- saving response for: %s  ---> DONE" \
                      % (info_station, req_cli, st_id))
            else:
                identifier += 1

        if input_dics['syngine']:
            dummy = 'syngine_waveform'
            syn_dirpath = os.path.join(
                target_path, 'syngine_%s' % input_dics['syngine_bg_model'])
            if not os.path.isdir(syn_dirpath):
                os.makedirs(syn_dirpath)
            if (not os.path.isfile(os.path.join(syn_dirpath, st_id)))\
                    or input_dics['force_waveform']:

                if input_dics['syngine_geocentric_lat']:
                    rcvlatitude = geocen_calc(float(st_avail[4]))
                    evlatitude = geocen_calc(event['latitude'])
                else:
                    rcvlatitude = float(st_avail[4])
                    evlatitude = event['latitude']

                if not event['focal_mechanism']:
                    syngine_momenttensor = None
                else:
                    syngine_momenttensor = event['focal_mechanism']

                # XXX some other arguments
                # sourcedoublecouple=None,
                # dt=None
                # kernelwidth=None
                # sourceforce=None
                # label=None
                req_syngine_component = st_avail[3][-1]
                if req_syngine_component == '1':
                    req_syngine_component = 'E'
                elif req_syngine_component == '2':
                    req_syngine_component = 'N'
                st_id = '%s.%s.%s.%s' % (st_avail[0],
                                         st_avail[1],
                                         st_avail[2],
                                         st_avail[3][:-1] + req_syngine_component)
                syn_st = client_syngine.get_waveforms(
                    model=input_dics['syngine_bg_model'],
                    receiverlatitude=rcvlatitude,
                    receiverlongitude=float(st_avail[5]),
                    networkcode=st_avail[0],
                    stationcode=st_avail[1],
                    sourcelatitude=evlatitude,
                    sourcelongitude=event['longitude'],
                    sourcedepthinmeters=float(event['depth'])*1000.,
                    origintime=event['datetime'],
                    components=req_syngine_component,
                    units=input_dics['syngine_units'],
                    sourcemomenttensor=syngine_momenttensor,
                    starttime=t_start,
                    endtime=t_end)[0]

                syn_st.stats.location = st_avail[2]
                syn_st.stats.channel = st_avail[3][:-1] + req_syngine_component
                syn_st.write(os.path.join(syn_dirpath, st_id),
                             format='mseed')

                identifier += 1000
                print('%s -- %s -- saving syngine for: %s  ---> DONE' \
                      % (info_station, req_cli, st_id))
            else:
                identifier += 1

        # if identifier in [0, 2, 3, 10, 11, 100]:
        #     raise Exception("CODE: %s will not be registered! (666)"
        #                     % identifier)
        t22 = datetime.now()
    except Exception as error:
        t22 = datetime.now()
        if len(st_avail) > 0:
            ee = '%s -- %s -- %s -- %s\n' % (req_cli, dummy, st_id, error)
        else:
            ee = '%s: There is no available station for this event.' % req_cli
        Exception_file = open(os.path.join(target_path,
                                           'info', 'exception'), 'at+')
        Exception_file.writelines(ee)
        Exception_file.close()

# ##################### fdsn_bulk_request ##################################


def fdsn_bulk_request(target_path, req_cli, input_dics):
    """
    send bulk request to FDSN
    :param target_path:
    :param req_cli:
    :param input_dics:
    :return:
    """
    print('\n[INFO] sending bulk request to: %s' % req_cli)

    client_fdsn = Client_fdsn(base_url=req_cli,
                              user=input_dics['username_fdsn'],
                              password=input_dics['password_fdsn'])

    bulk_list_fio = open(os.path.join(target_path, 'info',
                                      'bulkdata_list_%s' % req_cli), 'rb')
    bulk_list = pickle.load(bulk_list_fio)
    bulk_smgrs = client_fdsn.get_waveforms_bulk(bulk_list)
    print('[INFO] saving the retrieved waveforms from %s...' % req_cli)
    for bulk_st in bulk_smgrs:
        bulk_st.write(os.path.join(target_path, 'raw', '%s.%s.%s.%s'
                                   % (bulk_st.stats['network'],
                                      bulk_st.stats['station'],
                                      bulk_st.stats['location'],
                                      bulk_st.stats['channel'])),
                      'MSEED')

# ##################### update_sta_ev_file ##################################


def update_sta_ev_file(target_path, event):
    """
    update the station_event file based on already stored waveforms
    :param target_path:
    :param event:
    :return:
    """
    avail_arr = np.loadtxt(os.path.join(target_path, 'info',
                                        'availability.txt'),
                           delimiter=',', dtype=bytes, ndmin=2).astype(str)
    avail_arr = avail_arr.astype(object)
    sta_ev_add = os.path.join(target_path, 'info', 'station_event')
    sta_ev_fio = open(sta_ev_add, 'wt+')

    if not np.shape(avail_arr)[0] < 1:
        sta_ev_names = avail_arr[:, 0] + '.' + avail_arr[:, 1] + '.' + \
                       avail_arr[:, 2] + '.' + avail_arr[:, 3]

        sta_saved_path = glob.glob(
            os.path.join(target_path, 'raw', '*.*.*.*'))
        sta_saved_path.sort()
        sta_sorted = []

        for sta_sav_abs in sta_saved_path:
            try:
                sta_sav = os.path.basename(sta_sav_abs)
                sta_indx = np.where(sta_ev_names == sta_sav)[0][-1]
                sta_sorted.append(avail_arr[sta_indx])
            except:
                continue

        if len(np.shape(sta_sorted)) == 1:
            sta_sorted = np.reshape(sta_sorted, [1, len(sta_sorted)])
        if not np.shape(sta_sorted)[1] < 1:
            for sts in sta_sorted:
                sta_ev_line = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,\n' \
                              % (sts[0], sts[1], sts[2], sts[3], sts[4],
                                 sts[5], sts[6], sts[7], sts[8],
                                 event['event_id'], event['latitude'],
                                 event['longitude'], event['depth'],
                                 event['magnitude'], sts[10], sts[11], '10')
                sta_ev_fio.writelines(sta_ev_line)
    sta_ev_fio.close()

# -------------------------------- TRASH

# to shuffle the stations, we do not need it anymore as we
# parallelize it over the data-sources as well
# all_sta = []
# for req_cli in req_clients:
#     all_sta.append(stas_avail[stas_avail[:, -1] == req_cli])
# all_sta = np.array(list(roundrobin(all_sta)))

# from itertools import cycle, islice
# def roundrobin(iterables):
#     # Recipe credited to George Sakkis
#     pending = len(iterables)
#     nexts = cycle(iter(it).next for it in iterables)
#     while pending:
#         try:
#             for next in nexts:
#                 yield next()
#         except StopIteration:
#             pending -= 1
#     nexts = cycle(islice(nexts, pending))

# parallel_len_req_fdsn = range(0, len_req_fdsn)

# start = 0
# end = len_req_fdsn
# step = (end - start) / input_dics['req_np'] + 1

# jobs = []
# for index in xrange(input_dics['req_np']):
#     starti = start+index*step
#     endi = min(start+(index+1)*step, end)
#     p = multiprocessing.Process(target=FDSN_download_iter,
#                                 args=(i, starti, endi, dic, req_type,
#                                       len(events), events, add_event,
#                                       Sta_req, input_dics, client_fdsn))

#     jobs.append(p)
# for index in range(len(jobs)):
#     jobs[index].start()

# pp_flag = True
# while pp_flag:
#     for proc in jobs:
#         if proc.is_alive():
#             time.sleep(1)
#             pp_flag = True
#             break
#         else:
#             pp_flag = False
#     if not pp_flag:
#         print '\nAll the processes are finished...'

# len_par_grp = [parallel_len_req_fdsn[n:n+input_dics['req_np']] for n in
#                range(0, len(parallel_len_req_fdsn), input_dics['req_np'])]
# # ##################### FDSN_download_iter ##################################
#
#
# def FDSN_download_iter(i, starti, endi, dic, type, len_events, events,
#                        add_event, Sta_req, input, client_fdsn):
#     """
#     This function only iterates over FDSN_download_core,
#     this should be called by another program.
#     """
#     for j in range(starti, endi):
#         FDSN_download_core(i=i, j=j, dic=dic, type=type,
#                            len_events=len_events, events=events,
#                            add_event=add_event, Sta_req=Sta_req,
#                            input=input, client_fdsn=client_fdsn)



# ------------- required Python and obspy modules are imported in this part
from obspy.core import read
import os
import errno

from .utils.instrument_handler import instrument_correction
from .utils.resample_handler import resample_unit
from .utils.utility_codes import convert_to_sac
# -----------------------------------------------------------------------------

# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# ===================== YOU CAN CHANGE THE FOLLOWING FUNCTION =================
# >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# * IMPORTANT *
# The following function (process_unit) is in the waveform level.
# This means that you can write your process unit for one trace, and
# obspyDMT uses this function to pre-process all your waveforms,
# either right after retrieval or as a separate step:
# obspyDMT --datapath /your/dataset --local

# ========== process_unit has the following arguments:
# Use the following parameters to write your process_unit:
# 1. tr_add: address of one trace in your dataset. You can use that to
# read in the data.
# 2. target_path: address of the event that should be processed.
# 3. input_dics: dictionary that contains all the inputs.
# 4. staev_ar: an array that contains the following information:
# net, sta, loc, cha, station latitude, station longitude, station elevation,
# station depth


def process_unit(tr_add, target_path, input_dics, staev_ar):
    """
    processing unit, adjustable by the user
    :param tr_add: address of one trace in your dataset. You can use that to
    read in the data.
    :param target_path: address of the event that should be processed.
    :param input_dics: dictionary that contains all the inputs.
    :param staev_ar: an array that contains the following information:
           net, sta, loc, cha, station latitude, station longitude,
           station elevation, station depth
    :return:
    """
    # -------------- read the waveform, deal with gaps ------------------------
    # 1. read the waveform and create an obspy Stream object
    try:
        st = read(tr_add)
    except Exception as error:
        print('WARNING: %s' % error)
        return False
    # 2. in case that there are more than one waveform in a Stream (this can
    # happen due to some gaps in the waveforms) merge them.
    if len(st) > 1:
        try:
            st.merge(method=1, fill_value=0, interpolation_samples=0)
        except Exception as error:
            print('WARNING: %s' % error)
            return False
    # 3. Now, there is only one waveform, create a Trace
    tr = st[0]

    # -------------- waveform format ------------------------------------------
    # save_path is the address that will be used to save the processed data
    if input_dics['waveform_format'] == 'sac':
        save_path = os.path.join(target_path, 'renamed', "%s_%s.%s" % (tr.stats.starttime, tr.id, "SAC"))
    else:
        #"{staev_ar[9][:4]}/{staev_ar[9][4:6]}/{staev_ar[9][2:-2]}/{staev_ar[1][0]}/{staev_ar[0]}.{staev_ar[1]}/{staev_ar[9][2:-2]}.{staev_ar[1]}.{staev_ar[3]}"
        dir2save = os.path.join("/tmp/remove_me_KASRA/TEST/ami_data", staev_ar[9][:4], staev_ar[9][4:6], staev_ar[9][2:-2], staev_ar[1][0], "%s.%s" % (staev_ar[0], staev_ar[1]))
        if not os.path.isdir(dir2save):
            os.makedirs(dir2save)
        file_name = "%s.%s.%s" % (staev_ar[9][2:-2], staev_ar[1], staev_ar[3])
        save_path = os.path.join(dir2save, file_name)

    # -------------- path to save the processed waveform ----------------------
    # Before entering to the actual processing part of the code,
    # we define some paths to be used later:
    # you can adjust it as you want, here is just one example
    
    # If pathlib is installed, one can use: (suggested by ghraecakter)
    # pathlib.Path(os.path.join(target_path, 'processed')).mkdir(exist_ok=True)
    try:
        os.mkdir(os.path.join(target_path, 'renamed'))
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    if os.path.isfile(save_path) and (not input_dics['force_process']):
        return False

    # -------------- PROCESSING -----------------------------------------------
    # * resample the data
    # input_dics['sampling_rate'] determines the desired sampling rate
    if input_dics['sampling_rate']:
        print("resample %s from %s to %sHz" % (tr.id,
                                               tr.stats.sampling_rate,
                                               input_dics['sampling_rate']))
        tr = resample_unit(tr,
                           des_sr=input_dics['sampling_rate'],
                           resample_method=input_dics['resample_method'])

    # * apply instrument correction which consists of:
    # 1. removing the trend of the trace (remove_trend)
    # 2. remove the mean (zero_mean)
    # 3. taper (taper, taper_fraction [e.g., 0.05: 5%])
    # 4. apply pre-filter based on input_dics['pre_filt']
    # 5. apply instrument correction
    # all the above parameters are adjustable
    if input_dics['instrument_correction']:
        tr = instrument_correction(tr, target_path, save_path,
                                   input_dics['corr_unit'],
                                   input_dics['pre_filt'],
                                   input_dics['water_level'],
                                   zero_mean=True,
                                   taper=True,
                                   taper_fraction=0.05,
                                   remove_trend=True)

    # -------------- OUTPUT ---------------------------------------------------
    if not tr:
        pass
    elif input_dics['waveform_format'] == 'sac':
        tr = convert_to_sac(tr, save_path, staev_ar)
        tr.write(save_path, format='SAC')
    else:
        try:
            tr.write(save_path, format='mseed')
        except Exception as e:
            print('ERROR: %s -- %s' % (save_path, e))
